import importlib.metadata
import os
import sys
import json
import logging
from copy import deepcopy
import warnings
from functools import cache
from typing import Type, Any, Callable
from datetime import datetime
import numpy as np
import pandas as pd
from importlib.metadata import version

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
ALLOWED_CORRECTIONS = {'native', 'bonferroni', 'sidak', 'holm-sidak', 'holm', 'simes-hochberg', 'hommel', 'fdr_bh',
                       'fdr_by', 'fdr_tsbh', 'fdr_tsbky'}

logger = logging.getLogger('scoary.utils')

try:
    from ete3 import Tree as EteTree


    def print_tree(scoary_tree, label_to_gene: {str: bool}, label_to_trait: {str: bool}, show_label=True):
        if show_label:
            label_fn = lambda label: f'{int(label_to_gene[label])}{int(label_to_trait[label])}_{label}'
        else:
            label_fn = lambda label: f'{int(label_to_gene[label])}{int(label_to_trait[label])}'

        renamed_tree = scoary_tree.rename(label_fn)
        ete_tree = EteTree(renamed_tree.to_newick())
        print(ete_tree)

except ImportError as e:
    def print_tree(scoary_tree, label_to_gene: {str: bool}, label_to_trait: {str: bool}):
        raise ImportError('This function requires the ete3 library. Please install via "pip install ete3"')


@cache
def get_version() -> str:
    try:
        return version('scoary-2')
    except importlib.metadata.PackageNotFoundError:
        return 'development'


class NotSplittableError(Exception):
    pass


class NoTraitsLeftException(Exception):
    pass


def decode_unicode(string: str) -> str:
    return string.encode('utf-8').decode('unicode-escape')


def setup_outdir(outdir: str, input: dict) -> str:
    outdir = outdir.rstrip('/')
    assert not os.path.exists(outdir), f'ERROR: {outdir=} already exists!'
    os.makedirs(f'{outdir}/traits')
    os.makedirs(f'{outdir}/logs')
    os.makedirs(f'{outdir}/app')
    with open(f'{outdir}/logs/input.json', 'w') as f:
        json.dump(input, f, indent=4)
    return outdir


def setup_logging(logger: logging.Logger, path: str = None, print_info: bool = True, reset: bool = False):
    """
    Setup logging for Scoary

    :param logger: logging.logging.Logger
    :param path: if set, DEBUG and higher goes to log files
    :param print_info: if True, INFO and higher goes to stdout
    :param reset: if True: close and remove all file handlers. (Important for multiprocessing: removes locks!)
    :return:
    """
    if reset:
        while logger.handlers:
            handler = logger.handlers[0]
            handler.close()
            logger.removeHandler(handler)

    logger.setLevel(logging.DEBUG)

    if path is not None:
        # create logfile
        logfile = logging.FileHandler(path)
        logfile.setLevel(logging.DEBUG)
        logfile.setFormatter(logging.Formatter("%(asctime)s [%(name)s: %(levelname)s] %(message)s"))
        logger.addHandler(logfile)

    if print_info:
        # create streamhandler
        stdout = logging.StreamHandler()
        stdout.setLevel(logging.INFO)
        logger.addHandler(stdout)

    return logger


def ignore_warnings(warning: Type[Warning]):
    """
    Decorator to suppress warnings.

    Example:

    @ignore_warnings(warning=ConvergenceWarning)
    def some_function():
        # any produced ConvergenceWarnings will be suppressed

    :param warning: class of warning to be suppressed
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', warning)
                return function(*args, **kwargs)

        return wrapper

    return decorator


class RecursionLimit:
    """
    Context manager to temporarily change the recursion limit.

    Example:

    with RecursionLimit(10 ** 6):
        some_function_that_requires_lots_of_recursions()
    """
    new: int
    old: int

    def __init__(self, new_recursion_limit: int):
        self.new = new_recursion_limit

    def __enter__(self):
        self.old = sys.getrecursionlimit()
        logger.debug(f'Setting new recursion limit: {self.old} -> {self.new}')
        sys.setrecursionlimit(self.new)

    def __exit__(self, *args, **kwargs):
        logger.debug(f'Setting old recursion limit: {self.new} -> {self.old}')
        sys.setrecursionlimit(self.old)


def parse_correction(multiple_testing: str) -> (str, float):
    if ':' in multiple_testing:
        method, cutoff = multiple_testing.split(':', 1)
    else:
        method, cutoff = multiple_testing, 'inf'

    assert method in ALLOWED_CORRECTIONS, f'{multiple_testing=} must be in {ALLOWED_CORRECTIONS}'

    try:
        cutoff = float(cutoff)
    except ValueError:
        raise AssertionError(f'Error in {multiple_testing=}: {cutoff=} could not be converted to float')

    return method, cutoff


def is_int(string: str) -> bool:
    try:
        int(string)
        return True
    except ValueError:
        return False


def is_float(string: str) -> bool:
    try:
        float(string)
        return True
    except ValueError:
        return False


def split_into_parts(list_: list, n_parts: int) -> [list]:
    quotient, reminder = divmod(len(list_), n_parts)
    return [
        list_[i * quotient + min(i, reminder):(i + 1) * quotient + min(i + 1, reminder)]
        for i in range(n_parts)
    ]


def fisher_id(a, b, c, d):
    """
    Eight contingency tables always give the same pvalue: ['abcd', 'acbd', 'badc', 'bdac', 'cadb', 'cdab', 'dbca', 'dcba']

    Compute and save only one version.
    """
    return min((
        (a, b, c, d),
        (a, c, b, d),
        (b, a, d, c),
        (b, d, a, c),
        (c, a, d, b),
        (c, d, a, b),
        (d, b, c, a),
        (d, c, b, a)
    ))


def load_info_file(
        logger: logging.Logger,
        info_file: str,
        merge_col: str,
        expected_overlap_set: set = None,
        reference_file: str = None
) -> pd.DataFrame:
    """
    Load an info_file into pd.DataFrame:
        - Separator: tab ('\t')
        - Must have this header: {merge_col}\t{colname1}\t{colname2}...

    :param logger: instance of logging.Logger
    :param info_file: path to file
    :param merge_col: name of first column
    :param expected_overlap_set: a set of strings, some of which must occur in the index of info_file
    :param reference_file: path to reference file, just used for error messages
    :return: pd.DataFrame with merge_col as index
    """
    info_df = pd.read_csv(info_file, index_col=0, delimiter='\t')

    assert info_df.index.name == merge_col, \
        f'The file {info_file} is improperly formatted: The first column must be named "{merge_col}". ' \
        f'Current name: {info_df.index.name}. Remaining columns: {info_df.columns.tolist()}'

    if expected_overlap_set is not None:
        overlap_size = len(set.intersection(set(info_df.index), expected_overlap_set))
        if overlap_size == 0:
            logger.warning(f'The {merge_col}s in {info_file} do not match any {merge_col}s in {reference_file}')
        logger.debug(f'Loaded descriptions for {overlap_size} {merge_col}s')

    logger.debug(f'Loaded {merge_col} descriptions. columns={info_df.columns.tolist()}')
    assert not info_df.index.has_duplicates, \
        f'{info_file} contains duplicates: {info_df.index[info_df.index.duplicated()]}'
    return info_df


class MockCounter:
    """
    Imitate multiprocessing.Manager.Value / multiprocessing.managers.ValueProxy
    """

    def __init__(self):
        self._value: int = 0

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class MockLock:
    """
    Imitate multiprocessing.Manager.Lock / multiprocessing.managers.AcquirerProxy
    """

    def __init__(self):
        self._value = 0

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class AbstractNamespace:
    @classmethod
    def create_namespace(cls, ns, properties: {str: Any}):
        for name in cls.__dict__['__annotations__'].keys():
            setattr(ns, name, properties[name])
        return ns


def grasp_namespace(cls, ns):
    """
    This will copy the elements of the multiprocessing namespace into the "private" memory of the current process

    :param ns: multiprocessing.managers.Namespace
    :return: MockNameSpace
    """
    new_ns = cls()
    for name in cls.__dict__['__annotations__'].keys():
        value = getattr(ns, name)
        if name in ['lock', 'counter']:
            setattr(new_ns, name, value)
        else:
            setattr(new_ns, name, deepcopy(value))
    return new_ns


class AnalyzeTraitNamespace(AbstractNamespace):
    counter: MockCounter
    lock: MockLock
    outdir: str
    start_time: datetime
    genes_orig_df: pd.DataFrame
    genes_bool_df: pd.DataFrame
    gene_info_df: pd.DataFrame | None
    numeric_df: pd.DataFrame
    traits_df: pd.DataFrame
    trait_info_df: pd.DataFrame | None
    duplication_df: pd.DataFrame
    tree: object  #: ScoaryTree
    all_labels: set
    mt_f_method: str
    mt_f_cutoff: float
    max_genes: int
    worst_cutoff: None | float
    n_permut: int
    random_state: int
    pairwise: bool


class BinarizeTraitNamespace(AbstractNamespace):
    counter: MockCounter
    lock: MockLock
    outdir: str
    start_time: datetime
    numeric_df: pd.DataFrame
    random_state: int
    method: str
    alternative: str
    covariance_type: str
    cutoff: float
    random_state: int



from .progressbar import print_progress
from .utils import *
from .ScoaryTree import ScoaryTree
from .load_genes import load_genes
from .load_traits import load_traits
from .final_overview import create_final_overview
from .analyze_trait import analyze_trait, worker

logger = logging.getLogger('scoary')


def scoary(
        genes: str,
        traits: str,
        outdir: str,
        multiple_testing: str = 'bonferroni:0.999',
        worst_cutoff: float = None,
        max_genes: int = None,
        gene_info: str = None,
        trait_info: str = None,
        isolate_info: str = None,
        newicktree: str = None,
        pairwise: bool = True,
        n_permut: int = 500,
        restrict_to: str = None,
        ignore: str = None,
        n_cpus: int = 1,
        n_cpus_binarization: int = None,
        trait_data_type: str = 'binary:,',
        gene_data_type: str = 'gene-count:,',
        random_state: int = None,
        limit_traits: (int, int) = None,
        version: bool = False  # Dummy variable, only used to create docstring (see main function)
) -> None:
    """
    Scoary2: Associate genes with traits!

    :param genes: Path to gene presence/absence table: columns=isolates, rows=genes
    :param traits: Path to trait presence/absence table: columns=traits, rows=isolates
    :param outdir: Directory to place output files
    :param multiple_testing: "method:cutoff" for filtering genes after Fisher's test, where cutoff is a number that
    specifies the FWER and method is one of [native, bonferroni, sidak, holm-sidak, holm, simes-hochberg, hommel,
    fdr_bh, fdr_by,  fdr_tsbh, fdr_tsbky]. Alternatively, the method can be 'native': then, the cutoff targets the
    uncorrected p-value from Fisher's test.
    :param worst_cutoff: Drop traits if no gene with "worst" p-value lower than threshold. Recommended if
    dataset contains multiple species
    :param max_genes: Keep only n highest-scoring genes in Fisher's test. Recommended if dataset is big and contains
     multiple species; avoids waisting computational resources on traits that simply correlate with phylogeny
    :param gene_info: Path to file that describes genes: columns=arbitrary properties, rows=genes
    :param trait_info: Path to file that describes traits: columns=arbitrary properties, rows=traits
    :param isolate_info: Path to file that describes isolates: columns=arbitrary properties, rows=isolates
    :param newicktree: Path to a custom tree in Newick format
    :param pairwise: If False, only perform Fisher's test. If True, also perform pairwise comparisons
     algorithm.
    :param n_permut: Post-hoc label-switching test: perform N permutations of the phenotype by random label switching.
     Low p-values suggest that the effect is not merely lineage-specific.
    :param restrict_to: Comma-separated list of isolates to which to restrict this analysis
    :param ignore: Comma-separated list of isolates to be ignored for this analysis
    :param n_cpus: Number of CPUs that should be used. There is overhead in multiprocessing, so if the dataset is
    small, use n_cpus=1
    :param n_cpus_binarization: Number of CPUs that should be used for binarization. Default: one tenth of n_cpus
    :param trait_data_type: "<method>:<?cutoff>:<?covariance_type>:<?alternative>:<?delimiter>" How to read the traits
     table. Example: "gene-list:\\t" for OrthoFinder N0.tsv table
    :param gene_data_type: "<data_type>:<?delimiter>" How to read the genes table. Example: "gene-list:\\t" for
     OrthoFinder N0.tsv table
    :param random_state: Set a fixed seed for the random number generator
    :param limit_traits: Limit the analysis to traits n to m. Useful for debugging. Example: "(0, 10)"
    :param version: Print software version of Scoary2 and exit.
    """
    print(f'Welcome to Scoary2! ({get_version()})')

    # parse input, create outdir, setup logging
    trait_data_type = decode_unicode(trait_data_type)
    gene_data_type = decode_unicode(gene_data_type)
    if n_cpus_binarization is None:
        n_cpus_binarization = 1 + n_cpus // 10
    outdir = setup_outdir(outdir, input=locals())
    setup_logging(logger, f'{outdir}/logs/scoary-2.log')
    logger.debug(f'Scoary2 Version: {get_version()}')
    mt_f_method, mt_f_cutoff = parse_correction(multiple_testing)
    assert n_permut == 0 or n_permut >= 100, f'{n_permut=} must be at least 100.'

    # start
    start_time = datetime.now()

    # load traits data  (numeric_df may be None)
    logger.info('Loading traits...')
    numeric_df, traits_df = load_traits(
        traits=traits,
        trait_data_type=trait_data_type,
        restrict_to=restrict_to,
        ignore=ignore,
        n_cpus=n_cpus_binarization,
        random_state=random_state,
        outdir=outdir,
        limit_traits=limit_traits
    )

    # dynamically set recursion limit, should work for ~ 13'000 isolates
    _recursion_limit = max(1000, 100 + len(traits_df.index) ** 2)
    logger.debug(f'Setting recursion limit to {_recursion_limit}')
    sys.setrecursionlimit(_recursion_limit)

    if trait_info:
        logger.info('Loading trait info...')
        trait_info = load_info_file(
            logger=logger, info_file=trait_info, merge_col='Trait',
            expected_overlap_set=set(traits_df.columns), reference_file=traits
        )

    logger.info('Loading genes...')
    genes_orig_df, genes_bool_df = load_genes(
        genes,
        gene_data_type=gene_data_type,
        restrict_to=traits_df.index,
    )

    if gene_info:
        logger.info('Loading gene info...')
        gene_info = load_info_file(
            logger=logger, info_file=gene_info, merge_col='Gene',
            expected_overlap_set=set(genes_bool_df.index), reference_file=genes
        )

    if isolate_info:
        logger.info('Loading isolate info...')
        isolate_info = load_info_file(
            logger=logger, info_file=isolate_info, merge_col='Isolate',
            expected_overlap_set=set(genes_bool_df.columns), reference_file='placeholder'
        )

    # load phylogeny
    if newicktree is None:
        logger.info('Generating phylogenetic tree from gene presence-absence-matrix...')
        tree = ScoaryTree.from_presence_absence(genes_bool_df)
    else:
        logger.info('Loading phylogenetic tree from newick file...')
        with open(newicktree) as f:
            tree = ScoaryTree.from_newick(f.read())
    tree.write_newick(f'{outdir}/tree.nwk')

    all_labels = set(tree.labels())

    duplication_df = create_duplication_df(traits_df)

    logger.info('Finalizing setup...')
    if n_cpus == 1:
        ns, counter, lock = AnalyzeTraitNamespace(), MockCounter(), MockLock()
    else:
        from .init_multiprocessing import init, mp
        mgr, ns, counter, lock = init()

    ns = AnalyzeTraitNamespace.create_namespace(ns, {
        'start_time': datetime.now(),
        'counter': counter,
        'lock': lock,
        'outdir': outdir,
        'genes_orig_df': genes_orig_df,
        'genes_bool_df': genes_bool_df,
        'gene_info_df': gene_info,
        'numeric_df': numeric_df,
        'traits_df': traits_df,
        'trait_info_df': trait_info,
        'duplication_df': duplication_df,
        'tree': tree,
        'all_labels': all_labels,
        'mt_f_method': mt_f_method,
        'mt_f_cutoff': mt_f_cutoff,
        'max_genes': max_genes,
        'worst_cutoff': worst_cutoff,
        'n_permut': n_permut,
        'random_state': random_state,
        'pairwise': pairwise,
    })

    traits = traits_df.columns.to_list()

    logger.info('Start analyzing traits...')
    if n_cpus == 1:
        picking_start = datetime.now()
        trait_to_result = {trait: analyze_trait(trait, ns) for trait in traits}
    else:
        mp.freeze_support()
        queue = mgr.JoinableQueue()
        trait_to_result = mgr.dict()
        [queue.put(trait) for trait in traits]
        procs = [mp.Process(target=worker, args=(queue, ns, trait_to_result, i)) for i in range(n_cpus)]
        picking_start = datetime.now()
        [p.start() for p in procs]
        [p.join() for p in procs]
    picking_end = datetime.now()

    print_progress(
        len(ns.traits_df.columns), len(ns.traits_df.columns),
        message='COMPLETE!', start_time=ns.start_time, message_width=25,
        end='\n'
    )
    logging.info(f'Picking took {picking_end - picking_start}')
    try:
        summary_df = create_summary_df(trait_to_result)
        create_final_overview(summary_df, ns, isolate_info)

        logger.info('Complete success!')

    except NoTraitsLeftException as e:
        logger.info(str(e))

    logger.debug(f'Took {datetime.now() - start_time}')

    print(CITATION)


def create_summary_df(trait_to_result: {str: [dict | str | None]}) -> pd.DataFrame | None:
    """
    Turn trait_to_result into a pandas.DataFrame. Example:

             best_fisher_p  best_fisher_q  best_empirical_p  best_fq*ep
    Trait_1       0.574066   4.384058e-01          0.035964    0.035964
    Trait_2       0.432940   2.667931e-01          0.133866    0.133866
    Trait_3       0.194418   7.981206e-08          0.020979    0.691309

    :param trait_to_result: dictionary where keys are trait names and values are either dict|str|None
    :return: pandas.DataFrame
    """
    # res may contain: float, str or None
    # float: smallest empirical pvalue
    # str: duplicated trait -> name of trait
    # None: no gene was significant

    trait_to_result = {t: trait_to_result[r] if type(r) is str else r for t, r in
                       trait_to_result.items()}  # restore duplicates
    trait_to_result = {t: r for t, r in trait_to_result.items() if r is not None}  # remove Nones

    if len(trait_to_result) == 0:
        raise NoTraitsLeftException('No traits left after filtering')

    summary_df = pd.DataFrame(trait_to_result).T
    summary_df = summary_df.infer_objects()  # harmonize dtypes

    logger.debug(f'Created summary_df:\n{summary_df}')

    return summary_df


def create_duplication_df(traits_df: pd.DataFrame) -> pd.Series:
    """
    Returns a pd.Series that maps duplicated traits to the first occurrence
    """
    hash_df = pd.DataFrame(index=traits_df.columns)
    hash_df['hash'] = traits_df.apply(lambda x: hash(tuple(x)), axis=0)
    hash_df['is_duplicated'] = hash_df['hash'].duplicated(keep=False)
    hash_df['use_cache'] = hash_df['hash'].duplicated(keep='first')
    lookup_df = hash_df[hash_df['is_duplicated'] & ~hash_df['use_cache']].sort_values(by='hash')
    duplication_df = hash_df[hash_df['use_cache']]
    duplication_df = duplication_df['hash'].apply(
        func=lambda h: lookup_df.iloc[lookup_df.hash.searchsorted(h)].name
    )
    return duplication_df


CITATION = f'''
  ██████  ▄████▄   ▒█████   ▄▄▄       ██▀███ ▓██   ██▓   ░▒█████▒░ 
▒██    ▒ ▒██▀ ▀█  ▒██▒  ██ ▒████▄    ▓██   ██ ▒██  ██▒   ▒█▒   ██▒░
░ ▓██▄   ▒▓█    ▄ ▒██░  ██ ▒██  ▀█▄  ▓██ ░▄█   ▒██ ██░       ░█▀   
  ▒   ██ ▒▓▓▄ ▄██▒▒██   ██ ░██▄▄▄▄██ ▒██▀▀█▄   ░ ▐██▓░      ▄█     
▒██████▒   ▓███▀ ░░ ████▓▒░ ▓█   ▓██▒░██▓ ▒██▒   ██▒▓░   ░███████▒ 
▒ ▒▓▒ ▒ ░  ░▒ ▒  ░░ ▒░▒░▒░  ▒▒   ▓▒█░░ ▒▓ ░▒▓░  ██▒▒▒    ░▒▒  ░▒░  
░ ░▒  ░ ░  ░  ▒     ░ ▒ ▒░   ▒   ▒▒ ░  ░▒ ░ ▒░▓██ ░▒░     ░░   ▒░  
░  ░  ░  ░        ░ ░ ░ ▒    ░   ▒     ░░   ░ ▒ ▒ ░░           ░   
      ░  ░ ░          ░ ░        ░  ░   ░     ░ ░                  
         ░                                    ░ ░                  
                        Microbial Pan-GWAS


If you use Scoary2 ({get_version()}), please cite:
Roder, T. et al. Scoary2: Rapid association of phenotypic multi-omics 
data with microbial pan-genomes.
BioRxiv (2023) doi:10.1101/2023.04.19.537353.
'''.strip('\n')


def main():
    import sys, fire

    if '--version' in sys.argv:
        print(f'{get_version()}')
        exit(0)

    fire.Fire(scoary)


if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd


def _find_min(arr: np.ndarray, n_cols: int) -> (int, int):
    min_index = int(np.nanargmin(arr))
    x, y = (min_index // n_cols, min_index % n_cols)
    assert x > y
    return x, y


def _merge(arr: np.ndarray, node_list: [], cluster_sizes: [int], x: int, y: int):
    n_rows, n_cols = arr.shape
    assert n_rows == n_cols == len(node_list) == len(cluster_sizes)
    assert x > y

    # update arr
    for p in range(len(arr)):
        if p in (x, y):
            continue
        px1, px2 = (x, p) if p < x else (p, x)
        py1, py2 = (y, p) if p < y else (p, y)
        assert not (np.isnan(arr[px1, px2]) or np.isnan(arr[py1, py2]))

        # calculate mean difference
        arr[py1, py2] = (arr[px1, px2] * cluster_sizes[x] + arr[py1, py2] * cluster_sizes[y]) / (cluster_sizes[x] + cluster_sizes[y])  # row, col

    # remove row and col x
    arr = np.delete(arr, x, 0)
    arr = np.delete(arr, x, 1)

    # update labels
    new_label = [node_list[y], node_list[x]]
    del node_list[x]
    node_list[y] = new_label

    # update cluster_sizes
    cluster_sizes[y] = cluster_sizes[x] + cluster_sizes[y]
    del cluster_sizes[x]

    assert arr.shape == (n_rows - 1, n_cols - 1)
    assert len(node_list) == n_rows - 1
    return arr, node_list, cluster_sizes


def _upgma(arr: np.ndarray, node_list: [str]) -> []:
    # fill triu with nan
    arr[np.triu_indices(arr.shape[0], k=0)] = np.nan

    cluster_sizes = [1 for _ in range(len(node_list))]

    while len(node_list) > 1:
        n_rows, n_cols = arr.shape
        assert len(node_list) == n_rows == n_cols

        # find next columns to merge
        x, y = _find_min(arr, len(node_list))

        # merge columns and update labels
        arr, node_list, cluster_sizes = _merge(arr, node_list, cluster_sizes, x, y)

    assert len(node_list) == 1
    tree = node_list[0]

    return tree


def upgma(distances: pd.DataFrame) -> []:
    """
    Apply UPGMA (unweighted pair group method with arithmetic mean) algorithm.
    Returns unweighted tree in nested list form.

    Insipred by 'Creating a Phylogenetic Tree' by Oxford Academic (https://www.youtube.com/watch?v=09eD4A_HxVQ)

    :param distances: pandas.DataFrame: values: symmetric array; columns: tree labels
    :return: tree: nested list of strings
    """
    # split pandas.DataFrame into numpy.ndarray and labels
    labels: [] = [str(c) for c in distances.columns]
    arr: np.ndarray = distances.values.astype(float)

    # sanity checks
    assert len(set(labels)) == len(labels), f'labels are not unique! {labels=}'
    assert np.allclose(arr, arr.T, rtol=1e-05, atol=1e-08), f'arr is not symmetric! arr:\n{distances.to_string()}'
    assert arr.shape[0] == arr.shape[1]
    assert not np.isnan(arr).any(), 'Distance matrix contains nan'
    assert not np.isinf(arr).any(), 'Distance matrix contains inf'
    assert not np.any(distances < 0), 'Distances must be positive'

    return _upgma(arr=arr, node_list=labels)



import logging
from shutil import copy
import numpy as np
import pandas as pd
from scipy.cluster import hierarchy
from scipy.spatial.distance import cdist, squareform
import matplotlib as mpl

mpl.use('SVG')
# The SVG backend avoids this error message:
# ValueError: Image size of 700x165660 pixels is too large. It must be less than 2^16 in each direction.
# This allows for dendrograms with at least 20'000 traits

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.collections import PatchCollection, QuadMesh
from matplotlib.patches import Rectangle

from .utils import AnalyzeTraitNamespace, ROOT_DIR, RecursionLimit

logger = logging.getLogger('scoary.final_overview')


def plot_dendrogram(linkage_matrix: np.ndarray, labels: [str], ax: Axes) -> {}:
    with RecursionLimit(max(1000, len(linkage_matrix))):  # empirically tested for up to 20'000 traits
        dendrogram_params = hierarchy.dendrogram(
            linkage_matrix,
            orientation='left',
            labels=labels,
            no_labels=True,
            ax=ax
        )

    ax.set_xlim(left=1, right=0)
    ax.tick_params(
        axis='both', which='both',
        bottom=True, top=False, left=False, right=False,
        labelleft=False,
    )
    return dendrogram_params


def add_clickable_patches(patch_names, fig: Figure, ax: Axes, max_x: int):
    patches = [
        Rectangle(xy=(0, i * 10), width=max_x, height=10)
        for i in range(len(patch_names))
    ]

    # Create patch collection with specified colour/alpha
    pc = PatchCollection(
        patches,
        facecolors='white', alpha=1e-6,  # will result in 'opacity: 0'
        gid='clickable-patches',
        transform=ax.transData, figure=fig
    )

    # add urls -> this doesn't work
    # pc.set_urls([f'overview.html?trait={n}' for n in patch_names])

    fig.add_artist(pc)


def plot_manhattan_like(qvals: pd.DataFrame, fig: Figure, ax: Axes, column_defs: {str: dict}) -> [QuadMesh]:
    # determine y intervals: [0, 10, 20, ...]
    y = np.arange(start=5, stop=len(qvals.index) * 10, step=10)

    ax.set_xlim(left=0)

    max_x = 1.
    for i, (col, def_) in enumerate(column_defs.items()):
        if col not in qvals.columns:
            continue

        # create pcolormesh with logarithmic scale
        x = -np.log10(qvals[col])

        max_x = max(max_x, x.max())

        ax.scatter(
            x, y,
            marker=def_['marker'],
            color=def_['color']
        )

    # add y ticks and labels
    ax.yaxis.tick_right()
    ytick_locations = np.arange(start=5, stop=len(qvals) * 10, step=10)
    ax.set_yticks(ytick_locations, qvals.index)
    ax.tick_params(
        axis='both', which='both',
        bottom=False, top=False, left=False, right=False,
        labelbottom=False
    )

    # add grid
    ax.set_axisbelow(True)
    ax.set_xticks(ticks=np.arange(0, max_x + 1, 1), minor=True)
    ax.grid(visible=True, which='both', axis='x', linestyle='dashed')

    # add shape on top of colormesh and ticks that can be made clickable
    add_clickable_patches(qvals.index, fig, ax, max_x)


def create_final_overview(summary_df: pd.DataFrame, ns: AnalyzeTraitNamespace, isolate_info_df: pd.DataFrame = None):
    logger.debug('Adding preliminary summary.tsv...')
    summary_df.to_csv(f'{ns.outdir}/summary.tsv', sep='\t')

    logger.info('Adding isolate info...')
    if isolate_info_df is not None:
        isolate_info_df.to_csv(f'{ns.outdir}/isolate_info.tsv', sep='\t')

    logger.info('Copying files...')
    # copy files
    for file in ['overview.html', 'trait.html']:
        copy(src=f'{ROOT_DIR}/templates/{file}', dst=f'{ns.outdir}/{file}')
    for file in ['config.json', 'trait.js', 'trait.css', 'overview.js', 'overview.css', 'favicon.svg']:
        copy(src=f'{ROOT_DIR}/templates/{file}', dst=f'{ns.outdir}/app/{file}')

    summary_df_index = list(summary_df.index)
    if len(summary_df) > 1:
        logger.info('Calculating dendrogram linkage matrix...')
        # prepare data, create linkage_matrix
        pre_jaccard = ns.traits_df[summary_df.index].astype('float').T
        pre_jaccard = ((pre_jaccard.fillna(0.5) * 2) - 1).astype('int')  # False -> -1, NAN -> 0, True -> 1

        # whether class=0 or class=1 is arbitrary. Calculate both possibilities, take minimum
        d1 = cdist(pre_jaccard, pre_jaccard, metric='jaccard')
        d2 = cdist(pre_jaccard, 0 - pre_jaccard, metric='jaccard')
        jaccard_distance = np.minimum(d1, d2) * 2  # multiply by 2 to make maximal distance 1 again

        jaccard_df = pd.DataFrame(jaccard_distance, columns=pre_jaccard.index, index=pre_jaccard.index)
        del d1, d2, jaccard_distance
        linkage_matrix = hierarchy.linkage(
            squareform(jaccard_df, checks=True),
            method='average',
            optimal_ordering=True
        )

        # calculate plot proportions
        content_height = max(3., len(jaccard_df) / 6)  # height dependent on number of compounds
        whitespace_abs = 0.6  # absolute amount of whitespace
        total_height = content_height + whitespace_abs  # total plot height
        whitespace_rel = whitespace_abs / 3 / total_height  # relative amount of whitespace

        # create matplotlib figure
        plt.close()
        fig = plt.figure(figsize=(8, total_height))  # , dpi=4)
        gs = fig.add_gridspec(
            nrows=1, ncols=2, width_ratios=(2, 1),
            left=0.05, right=0.6, bottom=whitespace_rel * 2, top=1 - whitespace_rel,
            wspace=0, hspace=0
        )

        # get axes objects with shared y-axis
        ax_dendrogram = fig.add_subplot(gs[0, 0])
        ax_colorbar = fig.add_subplot(gs[0, 1], sharey=ax_dendrogram)

        logger.info('Plotting dendrogram...')
        # plot dendrogram
        dendrogram_params = plot_dendrogram(linkage_matrix, labels=jaccard_df.columns.values, ax=ax_dendrogram)

        # reindex summary_df according to order in dendrogram
        summary_df_index = dendrogram_params['ivl']
        summary_df = summary_df.reindex(summary_df_index)

        column_defs = {
            'best_fisher_q': {'marker': '$f$', 'color': 'tab:green'},
            'best_empirical_p': {'marker': '$e$', 'color': 'tab:purple'},
            'best_fq*ep': {'marker': '*', 'color': 'tab:red'}
        }
        cols = [col for col in column_defs.keys() if col in summary_df.columns]
        plot_manhattan_like(summary_df[cols], fig=fig, ax=ax_colorbar, column_defs=column_defs)

        # save plot
        plt.savefig(f'{ns.outdir}/overview_plot.svg', format='svg')
        plt.close()

    if ns.trait_info_df is not None:
        logger.info('Adding trait info...')
        summary_df = summary_df \
            .merge(ns.trait_info_df, left_index=True, right_index=True, how='left', copy=False) \
            .reindex(summary_df_index)  # merging destroys index order

    # save summary_df, ensure order matches plot
    logger.info('Adding summary.tsv...')
    summary_df.index.name = 'Trait'
    summary_df.to_csv(f'{ns.outdir}/summary.tsv', sep='\t')



import logging
from collections import defaultdict
import pandas as pd

logger = logging.getLogger('scoary.load_genes')


def filter_df(df: pd.DataFrame, restrict_to: [str] = None, ignore: [str] = None) -> pd.DataFrame:
    if ignore:
        ignore = set(ignore)
        missing = ignore.difference(set(df.columns))
        assert len(missing) == 0, f'Some strains in ignore were not found: {missing=}'
        df = df[[c for c in df.columns if c not in ignore]]

    if restrict_to is not None:
        restrict_to = set(restrict_to)
        have_cols = set(df.columns)
        cols_missing = restrict_to.difference(have_cols)
        assert len(cols_missing) == 0, f'Some strains in restrict_to were not found:' \
                                       f'\n{cols_missing=}' \
                                       f'\n{restrict_to=}' \
                                       f'\n{have_cols=}'
        cols_dropped = restrict_to.difference(set(df.columns))
        logger.debug(f'Cols kept: {list(restrict_to)}')
        logger.debug(f'Cols dropped: {list(cols_dropped)}')
        df = df[[c for c in df.columns if c in restrict_to]]

    return df


def load_gene_count_file(
        path: str,
        delimiter: str,
        restrict_to: [str] = None,
        ignore: [str] = None
) -> (pd.DataFrame, pd.DataFrame):
    """
    Load Roary-style gene count file with columns=strains and rows=genes

    :param path: Path to file
    :param delimiter: delimiter
    :param start_col: how many columns to skip
    :return: genes_df (DataFrame, dtype: bool); columns: strains; index: genes
    """
    count_df = pd.read_csv(path, delimiter=delimiter, index_col=0)

    # remove columns that are not in traits_df
    if restrict_to is not None or ignore is not None:
        count_df = filter_df(count_df, restrict_to, ignore)

    # sanity checks
    assert count_df.columns.is_unique, f'{path=}: columns not unique'
    assert count_df.index.is_unique, f'{path=}: index not unique'
    assert not count_df.isna().values.any(), f'{path=}: contains NaN'

    # add metadata
    count_df.attrs['content_type'] = 'gene-count'

    # convert to bool
    binary_df = count_df >= 1

    # remove core- and unique genes
    row_sums = binary_df.sum(axis=1)
    binary_df = binary_df[(row_sums != 0) & (row_sums != len(binary_df.columns))]

    logger.debug(f'Loaded gene-count-df:\n{binary_df}')
    return count_df, binary_df


def load_gene_list_file(
        path: str,
        restrict_to: [str] = None,
        ignore: [str] = None
) -> (pd.DataFrame, pd.DataFrame):
    """
    Load Orthofinder-style gene list file with columns=strains and rows=genes

    :param path: Path to file
    :return: genes_df (DataFrame, dtype: bool); columns: strains; index: genes
    """
    list_df = pd.read_csv(path, delimiter='\t', index_col=0, dtype=str)

    # remove columns that are not in traits_df
    if restrict_to is not None or ignore is not None:
        list_df = filter_df(list_df, restrict_to, ignore)

    # sanity checks
    assert list_df.columns.is_unique, f'{path=}: columns not unique'
    assert list_df.index.is_unique, f'{path=}: index not unique'

    # add metadata
    list_df.attrs['content_type'] = 'gene-list'

    # convert to bool
    binary_df = ~list_df.isna()

    # remove core- and unique genes
    row_sums = binary_df.sum(axis=1)
    binary_df = binary_df[(row_sums != 0) & (row_sums != len(binary_df.columns))]

    logger.debug(f'Loaded gene-list -df:\n{binary_df}')
    return list_df, binary_df


def parse_params(orig_params: str) -> (str, str):
    error_message = f"""
{orig_params=} is poorly formatted.
Must be '<data_type>:<?delimiter>'.
  Possible values for data_type:  {{'gene-count', 'gene-list'}}  (default: gene-count)
  Possible values for delimiter:  any single character, only relevant when data_type=gene-count  (default: ',')
""".strip()

    params = orig_params.lower().split(':')

    if len(params) == 1:
        data_type, delimiter = params[0], ','
    elif len(params) == 2:
        data_type, delimiter = params
    else:
        raise AssertionError(error_message)

    assert data_type in {'gene-count', 'gene-list'}, error_message

    return data_type, delimiter


def load_genes(
        genes: str,
        gene_data_type: str,
        restrict_to: [str] = None,
        ignore: [str] = None
) -> (pd.DataFrame, pd.DataFrame):
    """
    Load genes_df with columns=strains and rows=genes

    :param genes: Path to genes file
    :return: genes_df (DataFrame, dtype: bool); columns: strains; index: genes
    """
    data_type, delimiter = parse_params(gene_data_type)

    if data_type == 'gene-count':
        genes_orig_df, genes_bool_df = load_gene_count_file(genes, delimiter, restrict_to, ignore)
    elif data_type == 'gene-list':
        genes_orig_df, genes_bool_df = load_gene_list_file(genes, restrict_to, ignore)
    else:
        raise AssertionError(f'Programming error: {data_type=} must be gene-count or gene-list!')

    # ensure the index name is always the same
    genes_orig_df.index.name = 'Gene'
    genes_bool_df.index.name = 'Gene'

    return genes_orig_df, genes_bool_df



from __future__ import annotations

from functools import cached_property
from typing import Optional, Callable

import numpy as np
import pandas as pd
from scipy.spatial import distance

from .newick import parse_newick
from .upgma import upgma


class ScoaryTree:
    left: Optional[ScoaryTree] = None
    right: Optional[ScoaryTree] = None
    label: Optional[str] = None
    is_leaf: bool = False
    _values: Optional[np.ndarray] = None
    _prune = False

    def __init__(self, left: ScoaryTree = None, right: ScoaryTree = None, label: str = None):
        if left is None and right is None:
            self.is_leaf = True
            assert type(label) is str, f'A valid node has a label! {label=}'
            self.label = label
        else:
            self.is_leaf = False
            assert type(left) is ScoaryTree and type(
                right) is ScoaryTree, f'A valid tree has 0 or 2 children! {left=} {right=}'
            self.left = left
            self.right = right

    def __str__(self) -> str:
        return self._newick()

    def __repr__(self):
        return str(self)

    def _newick(self):
        return self.label if self.is_leaf else f"({self.left._newick()},{self.right._newick()})"

    def to_newick(self) -> str:
        return f'{self._newick()};'

    def write_newick(self, path: str):
        with open(path, 'w') as f:
            f.write(self.to_newick())

    def labels(self) -> [str]:
        if self.is_leaf:
            return [self.label]
        else:
            return self.left.labels() + self.right.labels()

    def uniquify(self, label_to_trait: {str: bool}):
        def uniquify(tree: ScoaryTree) -> str:
            if tree.is_leaf:
                return '1' if label_to_trait[tree.label] else '0'
            else:
                l, r = uniquify(tree.left), uniquify(tree.right)
                return f'({l}{r})' if l < r else f'({r}{l})'

        return uniquify(self)

    def copy(self):
        def copy(tree: ScoaryTree) -> ScoaryTree:
            if tree.is_leaf:
                return ScoaryTree(label=tree.label)
            else:
                return ScoaryTree(left=copy(tree.left), right=copy(tree.right))

        return copy(self)

    def prune(self, labels: [str]) -> ScoaryTree:
        n_labels_found = 0

        def prune(tree: ScoaryTree) -> Optional[ScoaryTree]:
            if tree.is_leaf:
                if tree.label in labels:
                    nonlocal n_labels_found
                    n_labels_found += 1
                    return ScoaryTree(label=tree.label)
                else:
                    return None
            else:
                left, right = prune(tree.left), prune(tree.right)
                if left and right:
                    return ScoaryTree(left=left, right=right)
                if left:
                    return left
                if right:
                    return right
                return None

        pruned_tree = prune(self)

        if n_labels_found != len(labels):
            missing = set(labels).difference(set(self.labels()))
            raise AssertionError(f'Pruning went wrong: did not find all labels in tree! '
                                 f'{n_labels_found=}; {missing=}; tree={self}')

        return pruned_tree

    def prune_nonrecursive(self, labels: [str]) -> ScoaryTree:
        if self.is_leaf:
            assert [self.label] == labels, f'Pruning went wrong. {[self.label]} != {labels}'
            return ScoaryTree(label=self.label)

        n_labels_found = 0

        root = ScoaryTree(left=self.left, right=self.right)

        stack = [(root, 'right'), (root, 'left')]

        while stack:
            current_parent, current_direction = stack[-1]
            current_node: ScoaryTree = getattr(current_parent, current_direction)

            if current_node.is_leaf:
                # current node is leaf

                this = ScoaryTree(label=current_node.label)

                if current_node.label in labels:
                    n_labels_found += 1
                else:
                    this._prune = True  # mark for pruning

                # append self to parent
                setattr(current_parent, current_direction, this)

                if current_direction == 'right':
                    # prune
                    current_parent.__prune()
                    stack.pop()

                    # found terminal node
                    # # GO UP UNTIL CAN GO RIGHT
                    while stack:
                        ancestor_node, ancestor_direction = stack[-1]
                        if ancestor_direction == 'right':
                            ancestor_node.__prune()
                            stack.pop()
                        else:
                            break

                    if not stack:
                        print(f'done\n{self}\n{root}')
                        break

                # pop left node -> go right next
                stack.pop()

            else:
                this = ScoaryTree(left=current_node.left, right=current_node.right)
                stack.extend([(this, 'right'), (this, 'left')])

                # append self to parent
                setattr(current_parent, current_direction, this)

        return root

    def copy_nonrecursive(self) -> ScoaryTree:
        return self.rename_nonrecursive(func=lambda label: label)

    def rename_nonrecursive(self, func: Callable):
        if self.is_leaf:
            return ScoaryTree(label=func(self.label))

        root = ScoaryTree(left=self.left, right=self.right)

        stack = [(root, 'right'), (root, 'left')]

        while stack:
            current_parent, current_direction = stack[-1]
            current_node: ScoaryTree = getattr(current_parent, current_direction)

            if current_node.is_leaf:
                # current node is leaf
                this = ScoaryTree(label=func(current_node.label))

                # append self to parent
                setattr(current_parent, current_direction, this)

                if current_direction == 'right':
                    # found terminal node
                    # # GO UP UNTIL CAN GO RIGHT
                    while stack and stack[-1][1] == 'right':
                        stack.pop()

                    if not stack:
                        print(f'done\n{self}\n{root}')
                        break

                # pop left node -> go right next
                stack.pop()

            else:
                this = ScoaryTree(left=current_node.left, right=current_node.right)
                stack.extend([(this, 'right'), (this, 'left')])

                # append self to parent
                setattr(current_parent, current_direction, this)

        return root

    def rename(self, func: Callable):
        """
        Apply a function to each leaf label.

        Only used for debugging. This recursive function could cause RecursionError for big trees.
        """

        def convert(tree: ScoaryTree) -> ScoaryTree:
            """recursive function"""
            if tree.is_leaf:
                return ScoaryTree(label=func(tree.label))
            else:
                return ScoaryTree(left=convert(tree.left), right=convert(tree.right))

        return convert(self)

    @classmethod
    def from_newick(cls, newick: str) -> ScoaryTree:
        list_tree = parse_newick(newick)
        return cls.from_list(list_tree)

    @classmethod
    def from_list(cls, tree: []) -> ScoaryTree:
        def convert(list_tree):
            """recursive function"""
            if type(list_tree) is str:
                return cls(label=list_tree)
            else:
                return cls(left=convert(list_tree[0]), right=convert(list_tree[1]))

        return convert(tree)

    @cached_property
    def to_list(self) -> []:
        def to_list(tree: ScoaryTree) -> str | []:
            if tree.is_leaf:
                return tree.label
            else:
                return [to_list(tree.left), to_list(tree.right)]

        return to_list(self)

    @classmethod
    def from_presence_absence(cls, genes_df: pd.DataFrame) -> ScoaryTree:
        distance_matrix = pd.DataFrame(distance.squareform(distance.pdist(genes_df.T, 'hamming')),
                                       columns=genes_df.columns)
        tree_as_list = upgma(distance_matrix)
        return cls.from_list(tree_as_list)

    def __prune(self):
        if self.left._prune and self.right._prune:
            self._prune = True
        elif self.left._prune:
            # become right
            self.label = self.right.label
            self.is_leaf = self.right.is_leaf
            self.left = self.right.left
            self.right = self.right.right
        elif self.right._prune:
            # become left
            self.label = self.left.label
            self.is_leaf = self.left.is_leaf
            self.right = self.left.right
            self.left = self.left.left



import os
from datetime import datetime, timedelta
from textwrap import shorten
import logging

# can os determine the terminal size?
try:
    n_cols = os.get_terminal_size().columns
    DYNAMIC_TERMINAL_WIDTH = True
    LINEBREAK_CHAR = '\r'
except Exception:
    DYNAMIC_TERMINAL_WIDTH = False
    LINEBREAK_CHAR = '\n'

# set function get_terminal_width depending on DYNAMIC_TERMINAL_WIDTH
if DYNAMIC_TERMINAL_WIDTH:
    def get_terminal_width(min_: int, default: int) -> int:
        return max(min_, os.get_terminal_size().columns)
else:
    def get_terminal_width(min_: int, default: int) -> int:
        return default


def stringify_timedelta(delta: timedelta) -> str:
    """
    Returns string 5 characters long.
    """
    d = delta.days
    h, rem = divmod(delta.seconds, 3600)
    m, s = divmod(rem, 60)
    if d:
        res = f'{d}d {h}h'
    elif h:
        res = f'{h}h {m}m'
    elif m:
        res = f'{m}m {s}s'
    else:
        res = f'{s}s'
    res = shorten(res, width=5, placeholder='')
    return f'{res:>5s}' if res else f'>999d'


def print_progress(
        i: int,
        n: int,
        message: str,
        start_time: datetime,
        message_width: int = 40,
        default_width: int = 100,
        sep: str = ' | ',
        end: str = LINEBREAK_CHAR
) -> None:
    message = f"{shorten(message, width=message_width, placeholder='...'):{message_width}s}"
    assert len(message) == message_width

    n = max(1, n)
    i_safe = min(max(1, i), n)
    time_left = stringify_timedelta((datetime.now() - start_time) / i_safe * (n - i_safe))  # 5 chars
    percentage = f"{f'{i / n:.0%}':>4s}"  # 4 chars

    width_total = get_terminal_width(min_=message_width + len(sep) * 2 + 20, default=default_width)

    text = f'{percentage}{sep}{time_left}{sep}{message}'
    len_progressbar = width_total - len(text)
    n_bars = len_progressbar - 3  # because of '[] '

    res = f"[{'=' * round(i / n * n_bars):{n_bars}}] {text}"

    if not len(res) == width_total:
        logging.warning('Failed to print progressbar!')

    print(res, end=end)



import logging
import queue
from datetime import datetime
from typing import Callable

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from pandas._libs.parsers import STR_NA_VALUES
from sklearn.exceptions import ConvergenceWarning
from sklearn.mixture import GaussianMixture

from .progressbar import print_progress
from .utils import setup_logging, is_float, ignore_warnings, BinarizeTraitNamespace, MockLock, MockCounter, \
    grasp_namespace, NotSplittableError
from queue import Empty

logger = logging.getLogger('scoary.load_traits')
STR_NA_VALUES.update(['-', '.'])  # compatibility with Scoary 1


def filter_df(df: pd.DataFrame, restrict_to: str = None, ignore: str = None) -> pd.DataFrame:
    if ignore:
        ignore = set(ignore.split(','))
        missing = ignore.difference(set(df.index))
        assert len(missing) == 0, f'Some strains in ignore were not found: {missing=}'
        df = df[[i not in ignore for i in df.index]]

    if restrict_to:
        restrict_to = set(restrict_to.split(','))
        missing = restrict_to.difference(set(df.index))
        assert len(missing) == 0, f'Some strains in restrict_to were not found: {missing=}'
        df = df[[i in restrict_to for i in df.index]]

    return df


def switch_labels(classes: pd.Series, means: np.array) -> pd.Series:
    if means[0] > means[1]:
        return classes
    else:
        return ~classes


def load_binary(traits: str, delimiter: str, restrict_to: str = None, ignore: str = None,
                limit_traits: (int, int) = None):
    logger.debug(f'Loading binary traits: {traits=} {delimiter=}')
    traits_df = pd.read_csv(traits, delimiter=delimiter, index_col=0, na_values=STR_NA_VALUES)

    if limit_traits is not None:
        assert len(limit_traits) == 2, f'{limit_traits=} is poorly formatted: must be (int, int).'
        l1, l2 = limit_traits
        assert type(l1) is type(l2) is int, f'{limit_traits=} is poorly formatted: must be (int, int).'
        traits_df = traits_df[traits_df.columns[l1:l2]]

    filter_df(traits_df, restrict_to, ignore)
    assert traits_df.columns.is_unique, f'{traits=}: columns are not unique'
    assert traits_df.index.is_unique, f'{traits=}: index not unique'

    # convert to bool
    traits_df = traits_df.astype('boolean')

    n_nan = int(traits_df.isna().sum().sum())
    if n_nan:
        logger.debug(f'Found {n_nan} NaN in {traits=}')

    traits_df.attrs['binarization_method'] = 'none'
    traits_df.attrs['binarization_info'] = 'none'

    traits_df.index.name = 'Isolate'

    logger.debug(f'Loaded binary traits_df:\n{traits_df}')

    return traits_df


def load_numeric(traits: str, delimiter: str, restrict_to: str = None, ignore: str = None,
                 limit_traits: (int, int) = None):
    logger.debug(f'Loading numeric traits: {traits=} {delimiter=}')
    numeric_df = pd.read_csv(traits, delimiter=delimiter, index_col=0, na_values=STR_NA_VALUES)

    if limit_traits is not None:
        assert len(limit_traits) == 2, f'{limit_traits=} is poorly formatted: must be (int, int).'
        l1, l2 = limit_traits
        assert type(l1) is type(l2) is int, f'{limit_traits=} is poorly formatted: must be (int, int).'
        numeric_df = numeric_df[numeric_df.columns[l1:l2]]

    assert numeric_df.columns.is_unique, f'{traits=}: columns are not unique'
    assert numeric_df.index.is_unique, f'{traits=}: index not unique'

    n_nan = int(numeric_df.isna().sum().sum())
    if n_nan:
        logger.debug(f'Found {n_nan} NaN in {traits=}')

    numeric_df = filter_df(numeric_df, restrict_to, ignore)

    numeric_df.index.name = 'Isolate'

    logger.debug(f'Loaded numeric_df:\n{numeric_df}')

    return numeric_df


def apply_kmeans(kmeans: KMeans, data: pd.Series) -> (pd.Series, dict):
    nan_free_data = data[~np.isnan(data.values)]
    values = nan_free_data.values.reshape(-1, 1)
    fit = kmeans.fit(values)
    labels_ = fit.labels_.astype(bool)

    switch_labels(labels_, kmeans.cluster_centers_.flatten())

    # insert_positions = [v - i for i, v in enumerate(np.where(nan_vector)[0])]
    label_to_class = dict(zip(nan_free_data.index, labels_))
    return pd.array([label_to_class.get(l, pd.NA) for l in data.index], dtype="boolean"), \
        {'method': 'kmeans', 'cutoff': np.mean(fit.cluster_centers_)}


def classify_gm(val: [float, float], certainty_cutoff: float):
    if val[0] >= certainty_cutoff:  # probability of class 1
        return True
    elif val[1] >= certainty_cutoff:  # probability of class 2
        return False
    else:
        return pd.NA


def generate_extract_covariances(covariance_type: str) -> Callable:
    def create_tuple(c1: float, c2: float) -> (float, float):
        assert type(c1) is np.float64 and type(c2) is np.float64, \
            f'Failed to read covariances! {covariance_type=} {c1=} {c2=} {type(c1)=} {type(c2)=}'
        return c1, c2

    if covariance_type == 'diag':
        return lambda covariances_: create_tuple(covariances_[0][0], covariances_[1][0])
    elif covariance_type == 'tied':
        return lambda covariances_: create_tuple(covariances_[0][0], covariances_[0][0])
    elif covariance_type == 'full':
        return lambda covariances_: create_tuple(covariances_[0][0][0], covariances_[1][0][0])
    elif covariance_type == 'spherical':
        return lambda covariances_: create_tuple(*covariances_)
    else:
        raise AssertionError('Failed to extract covariances!')


@ignore_warnings(warning=ConvergenceWarning)
def apply_gm(gm: GaussianMixture, data: pd.Series, certainty_cutoff: float, extract_covariances: Callable
             ) -> (pd.Series, dict):
    nan_free_data = data[~np.isnan(data.values)]
    values = nan_free_data.values.reshape(-1, 1)
    fit = gm.fit(values)
    if not fit.converged_:
        raise NotSplittableError(f'GaussianMixture did not converge: {data.name}')
    densities = fit.predict_proba(values)
    label_to_class = {l: classify_gm(d, certainty_cutoff) for l, d in zip(nan_free_data.index, densities)}
    result = pd.array([label_to_class.get(l, pd.NA) for l in data.index], dtype="boolean")

    n_pos = result.sum()
    n_neg = (~result).sum()

    if n_pos < 1 or n_neg < 1:
        raise NotSplittableError(f'GaussianMixture failed to split {data.name}')

    result = switch_labels(result, gm.means_.flatten())

    metadata = {
        'method': 'gaussian', 'success': True,
        'means': [m[0] for m in gm.means_],
        'weights': list(gm.weights_),
        'covariances': extract_covariances(gm.covariances_),
    }

    return result, metadata


def parse_params(orig_params: str):
    error_message = f"""
{orig_params=} is poorly formatted.
Must be 'kmeans', 'gaussian' or '<method>:<?cutoff>:<?covariance_type>:<?alternative>:<?delimiter>'.
  Possible values for method:          {{'gaussian', 'kmeans'}}
  Possible values for delimiter:       any single character  (default: ',')
  Possible values for cutoff:          .5 <= cutoff < 1  (default: 0.85)
  Possible values for covariance_type: {{'tied', 'full', 'diag', 'spherical'}}  (default: tied)
  Possible values for alternative:     {{'skip', 'kmeans'}}  (default: skip)
""".strip()

    delimiter, cutoff, covariance_type, alternative = None, None, None, None

    params = orig_params.lower().split(':')
    method = params.pop(0)
    assert method in {'binary', 'kmeans', 'gaussian'}, error_message

    for param in params:
        if param in {'full', 'tied', 'diag', 'spherical'}:
            assert covariance_type is None, error_message
            covariance_type = param
            continue
        elif param in {'skip', 'kmeans'}:
            assert alternative is None, error_message
            alternative = param
            continue
        elif is_float(param):
            assert cutoff is None, error_message
            cutoff = float(param)
        elif len(param) == 1:
            assert delimiter is None, error_message
            delimiter = param
            continue
        else:
            raise AssertionError(error_message)

    if delimiter is None:
        delimiter = ','

    if cutoff is None:
        cutoff = 0.85

    if covariance_type is None:
        covariance_type = 'tied'

    if alternative is None:
        alternative = 'skip'

    assert 0.5 <= cutoff < 1, f'The cutoff must be between 0.5 and 1! {cutoff=}; {orig_params=}'

    return method, delimiter, cutoff, covariance_type, alternative


def print_status(ns, trait, proc_id):
    with ns.lock:
        ns.counter.value += 1
        message = trait if proc_id is None else f'P{proc_id} | {trait}'
        print_progress(
            ns.counter.value, len(ns.numeric_df.columns),
            message=message, start_time=ns.start_time, message_width=25
        )


def fn_km(kmeans, gm, ns, trait, proc_id, container_binarized_traits, container_binarization_info, extract_covariances):
    print_status(ns, trait, proc_id)
    logger.debug(f'Binarizing {trait=} using kmeans')
    binarized_data, metadata = apply_kmeans(kmeans, ns.numeric_df[trait])
    container_binarized_traits[trait] = binarized_data
    container_binarization_info[trait] = metadata


def fn_gm(kmeans, gm, ns, trait, proc_id, container_binarized_traits, container_binarization_info, extract_covariances):
    print_status(ns, trait, proc_id)
    logger.debug(f'Binarizing {trait=} using gaussian mixture.')
    try:
        binarized_data, metadata = apply_gm(gm, ns.numeric_df[trait], certainty_cutoff=ns.cutoff,
                                            extract_covariances=extract_covariances)
        container_binarized_traits[trait] = binarized_data
        container_binarization_info[trait] = metadata

    except NotSplittableError as e:
        if ns.alternative == 'skip':
            logger.debug(f"Skipping trait '{trait}': {e}.")
            container_binarization_info[trait] = {'method': 'gaussian', 'success': False}
        elif ns.alternative == 'kmeans':
            logger.debug(f"Using kmeans for trait '{trait}': {e}")
            binarized_data, metadata = apply_kmeans(kmeans, ns.numeric_df[trait])
            container_binarized_traits[trait] = binarized_data
            container_binarization_info[trait] = metadata
        else:
            raise AssertionError(f'Programming error: {ns.alternative=} must be skip or kmeans!')


def worker(
        ns: BinarizeTraitNamespace,
        container_binarized_traits: dict = None,
        container_binarization_info: dict = None,
        queue: queue.Queue = None,
        proc_id: int = None
):
    if proc_id is None:
        # single-CPU: create vessels
        container_binarized_traits = {}
        container_binarization_info = {}
    else:
        # multiprocessing: set up logging for this process
        logger = setup_logging(
            logger=logging.getLogger('scoary'),
            path=f'{ns.outdir}/logs/scoary-2_proc{proc_id}.log',
            print_info=False,
            reset=True
        )
        logger.info(f'Setting up binarization worker {proc_id}')

    # copy all data to this Python interpreter
    new_ns = grasp_namespace(BinarizeTraitNamespace, ns)
    del ns

    kmeans = KMeans(n_init='auto', n_clusters=2, random_state=new_ns.random_state)
    gm = GaussianMixture(n_components=2, random_state=new_ns.random_state, covariance_type=new_ns.covariance_type)

    if new_ns.method == 'kmeans':
        extract_covariances = None
        fn = fn_km
    elif new_ns.method == 'gaussian':
        extract_covariances = generate_extract_covariances(new_ns.covariance_type)
        fn = fn_gm
    else:
        raise AssertionError(f'Programming error: {new_ns.method=} must be kmeans or gaussian!')

    if queue is None:
        for trait in new_ns.numeric_df.columns:
            fn(kmeans, gm, new_ns, trait, proc_id, container_binarized_traits, container_binarization_info,
               extract_covariances)
        return container_binarized_traits, container_binarization_info
    else:
        while True:
            try:
                trait = queue.get_nowait()
            except Empty:
                break  # completely done

            fn(kmeans, gm, new_ns, trait, proc_id, container_binarized_traits, container_binarization_info,
               extract_covariances)
            queue.task_done()


def binarize(
        numeric_df: pd.DataFrame,
        method: str,
        random_state: int | None,
        n_cpus: int,
        cutoff: float,
        covariance_type: str,
        alternative: str,
        outdir: str = None
) -> pd.DataFrame:
    """
    Convert numeric data to binary data.

    :param traits: Path to traits file
    :param outdir: path to outdir
    :return: numeric_df (DataFrame, dtype: float), traits_df (DataFrame, dtype: bool); columns: trait_names; index: strains
    """
    if n_cpus == 1:
        ns, lock, counter = BinarizeTraitNamespace(), MockLock(), MockCounter()
    else:
        from .init_multiprocessing import init, mp
        mgr, ns, counter, lock = init()

    logger.info(f'Binarizing traits: {method=} {alternative=} {cutoff=} {covariance_type=} {n_cpus=} {random_state=}')

    ns = BinarizeTraitNamespace.create_namespace(ns, {
        'start_time': datetime.now(),
        'lock': lock,
        'outdir': outdir,
        'counter': counter,
        'numeric_df': numeric_df,
        'method': method,
        'alternative': alternative,
        'covariance_type': covariance_type,
        'cutoff': cutoff,
        'random_state': random_state,
    })

    if n_cpus == 1:
        container_binarized_traits, container_binarization_info = worker(ns)
    else:
        mp.freeze_support()
        queue = mgr.JoinableQueue()
        container_binarized_traits = mgr.dict()
        container_binarization_info = mgr.dict()
        [queue.put(trait) for trait in numeric_df.columns]

        procs = [mp.Process(target=worker, args=(
            ns, container_binarized_traits, container_binarization_info, queue, i
        )) for i in range(n_cpus)]
        [p.start() for p in procs]
        [p.join() for p in procs]
        container_binarized_traits = dict(container_binarized_traits)
        container_binarization_info = dict(container_binarization_info)

    traits_df = pd.DataFrame.from_dict(container_binarized_traits)
    traits_df.index = numeric_df.index
    traits_df = traits_df[[c for c in numeric_df.columns if c in traits_df.columns]]

    if len(traits_df.columns) == 0:
        raise AssertionError(f'Failed to binarize traits!')

    # add metadata to DataFrame
    traits_df.attrs['binarization_method'] = method
    traits_df.attrs['binarization_info'] = container_binarization_info

    if outdir:
        traits_df.to_csv(f'{outdir}/binarized_traits.tsv', sep='\t')

    logger.debug(f'Binarized traits:\n{traits_df}')

    return traits_df


def load_traits(
        traits: str,
        trait_data_type: str = 'binary:,',
        restrict_to: str = None,
        ignore: str = None,
        n_cpus: int = 1,
        random_state: int = None,
        outdir: str = None,
        limit_traits: (int, int) = None
) -> (pd.DataFrame, pd.DataFrame):
    method, delimiter, cutoff, covariance_type, alternative = parse_params(orig_params=trait_data_type)

    if method == 'binary':
        numeric_df = None
        traits_df = load_binary(
            traits=traits, delimiter=delimiter,
            restrict_to=restrict_to, ignore=ignore,
            limit_traits=limit_traits
        )

    else:
        numeric_df = load_numeric(
            traits=traits, delimiter=delimiter,
            restrict_to=restrict_to, ignore=ignore,
            limit_traits=limit_traits
        )
        logger.info('Binarizing traits...')
        traits_df = binarize(
            numeric_df=numeric_df, method=method, random_state=random_state,
            cutoff=cutoff, covariance_type=covariance_type, alternative=alternative, n_cpus=n_cpus,
            outdir=outdir
        )
        print_progress(
            i=len(numeric_df.columns), n=len(numeric_df.columns),
            message='COMPLETE!', start_time=datetime.now(), message_width=25,
            end='\n'
        )

    return numeric_df, traits_df


if __name__ == '__main__':
    import fire

    fire.Fire(binarize)



from functools import cache

from numba import njit
import numpy as np
import pandas as pd
from scipy.stats import binom_test
from scipy.stats import binomtest

from .ScoaryTree import ScoaryTree


def pick(
        tree: [],
        label_to_trait_a: {str: bool},
        trait_b_df: pd.DataFrame,
        calc_pvals: bool = True
) -> (np.array,):
    """
    Traverse the tree and perform pair picking

    :param tree: Tree in list form
    :param label_to_trait_a: maps each label of the tree to whether it has trait a
    :param trait_b_df: DataFrame (dtype:bool); columns: labels of the tree; rows: whether trait b is present
    :param calc_pvals: If False, binomial test will not be applied and best/worst will be None
    :return: (max_contr, max_suppo, max_oppos, best, worst) if calc_pvals else (max_contr, max_suppo, max_oppos)
    """

    assert not trait_b_df.isna().values.any()

    def _pick(left_label, right_label):
        # follow tree until terminal node
        if type(left_label) is not str:
            left = _pick(left_label[0], left_label[1])
        if type(right_label) is not str:
            right = _pick(right_label[0], right_label[1])

        # only load new leafs when needed for combination (safe RAM)
        if type(left_label) is str:
            left = init_leaf(
                trait_a=label_to_trait_a[left_label],
                trait_b_list=trait_b_df[left_label].to_numpy(dtype='bool')
            )
        if type(right_label) is str:
            right = init_leaf(
                trait_a=label_to_trait_a[right_label],
                trait_b_list=trait_b_df[right_label].to_numpy(dtype='bool')
            )

        combined = combine_branches(left, right)

        return combined

    values = _pick(tree[0], tree[1])

    max_contr = values[:, 0, :].max(axis=1)
    max_suppo = values[:, 1, :].max(axis=1)
    max_oppos = values[:, 2, :].max(axis=1)

    if not calc_pvals:
        return max_contr, max_suppo, max_oppos

    best, worst = apply_binomtest(max_contr, max_suppo, max_oppos)

    return max_contr, max_suppo, max_oppos, best, worst


def pick_single(
        tree: [],
        label_to_trait_a: {str: bool},
        label_to_trait_b: {str: bool},
        calc_pvals: bool = True
) -> {str: int | float}:
    res = pick(
        tree=tree,
        label_to_trait_a=label_to_trait_a,
        trait_b_df=pd.DataFrame([label_to_trait_b]),
        calc_pvals=calc_pvals
    )
    return dict(zip(
        ['max_contrasting_pairs', 'max_supporting_pairs', 'max_opposing_pairs', 'best_pval', 'worst_pval'],
        [v[0] for v in res]
    ))


def pick_nonrecursive(
        tree: [],
        label_to_trait_a: {str: bool},
        trait_b_df: pd.DataFrame,
        calc_pvals: bool = True
) -> (np.array, np.array, np.array, np.array, np.array):
    if tree.is_leaf:
        return init_leaf(
            trait_a=label_to_trait_a[tree.label],
            trait_b_list=trait_b_df[tree.label].to_numpy(dtype='bool')
        )

    stack = [[tree, 'right'], [tree, 'left']]

    while stack:
        current_parent, current_direction = stack[-1]
        current_node: ScoaryTree = getattr(current_parent, current_direction)

        if current_node.is_leaf:
            # current node is leaf
            this = init_leaf(
                trait_a=label_to_trait_a[current_node.label],
                trait_b_list=trait_b_df[current_node.label].to_numpy(dtype='bool')
            )

            # append data to parent
            current_node._values = this

            if current_direction == 'right':
                # found terminal node
                # # GO UP UNTIL CAN GO RIGHT
                while stack and stack[-1][1] == 'right':
                    ancestor_tree, ancestor_direction = stack.pop()
                    ancestor_tree._values = combine_branches(
                        ancestor_tree.left._values,
                        ancestor_tree.right._values
                    )
                    ancestor_tree.left._values = None
                    ancestor_tree.right._values = None

                if not stack:
                    # arrived at root node
                    break

            # pop left node -> go right next
            stack.pop()

        else:
            stack.extend([(current_node, 'right'), (current_node, 'left')])

    values = tree._values
    tree._values = None

    max_contr = values[:, 0, :].max(axis=1)
    max_suppo = values[:, 1, :].max(axis=1)
    max_oppos = values[:, 2, :].max(axis=1)

    if not calc_pvals:
        return max_contr, max_suppo, max_oppos

    best, worst = apply_binomtest(max_contr, max_suppo, max_oppos)

    return max_contr, max_suppo, max_oppos, best, worst


@cache
def _binomtest(k: int, n: int) -> float:
    # caching this function increases speed ~ 40x
    return binomtest(k=k, n=n).pvalue


def apply_binomtest(max_contr, max_suppo, max_oppos):
    n_traits = max_contr.shape[0]
    result = np.empty(shape=(2, n_traits), dtype='float')

    for i in range(n_traits):
        b = _binomtest(max_suppo[i], n=max_contr[i])
        w = _binomtest(max_oppos[i], n=max_contr[i])

        if b < w:
            result[0][i] = b
            result[1][i] = w
        else:
            result[0][i] = w
            result[1][i] = b
    return result


# selecting:values[<TRAITS>, <3 TYPES OF PAIRINGS>, <5 COMBINATIONS>]
# selecting:values[<TRAITS>, <0: max; 1: supporting; 2: opposing>, <0: 11; 1: 10; 2: 01; 3: 00; 4: nf>]

# values[n, 0, :] -> all max contrasting pairs for trait n
# values[n, 1, :] -> all max supporting pairs for trait n
# values[n, 2, :] -> all max opposing pairs for trait n

# values[n, 0, 0] -> max supporting pairs for trait n if condition '11' is added
# values[n, 0, 1] -> max supporting pairs for trait n if condition '10' is added
# values[n, 0, 2] -> max supporting pairs for trait n if condition '01' is added
# values[n, 0, 3] -> max supporting pairs for trait n if condition '00' is added
# values[n, 0, 4] -> max supporting pairs for trait n if condition 'nf' is added


@njit('int64[:, ::3, ::5](boolean, boolean[:])',
      cache=True, nogil=True, boundscheck=False, parallel=False)  # prange not better
def init_leaf(trait_a: bool, trait_b_list: np.array) -> np.array:
    n_traits = trait_b_list.shape[0]

    values = np.full(shape=(n_traits, 3, 5), fill_value=-1, dtype='int')
    if trait_a:
        values[:, :, 0][trait_b_list] = 0
        values[:, :, 1][~trait_b_list] = 0

    else:
        values[:, :, 2][trait_b_list] = 0
        values[:, :, 3][~trait_b_list] = 0

    return values


@njit('int64[::3, ::5], int64[::3, ::5]',
      cache=True, nogil=True, boundscheck=False, parallel=False)  # parallel kills performance
def calculate_max_nofree(left: np.array, right: np.array):
    values = np.full(shape=(3, 5), fill_value=-1, dtype='int')

    if left[0][4] > -1 and right[0][4] > -1:  # nf vs nf
        values[0][0] = left[0][4] + right[0][4]
        values[1][0] = left[1][4] + right[1][4]
        values[2][0] = left[2][4] + right[2][4]

    if left[0][0] > -1 and right[0][3] > -1:  # 11 vs 00
        values[0][1] = left[0][0] + right[0][3] + 1
        values[1][1] = left[1][0] + right[1][3] + 1
        values[2][1] = left[2][0] + right[2][3]

    if left[0][3] > -1 and right[0][0] > -1:  # 00 vs 11
        values[0][2] = left[0][3] + right[0][0] + 1
        values[1][2] = left[1][3] + right[1][0] + 1
        values[2][2] = left[2][3] + right[2][0]

    if left[0][1] > -1 and right[0][2] > -1:  # 10 vs 01
        values[0][3] = left[0][1] + right[0][2] + 1
        values[1][3] = left[1][1] + right[1][2]
        values[2][3] = left[2][1] + right[2][2] + 1

    if left[0][2] > -1 and right[0][1] > -1:  # 01 vs 10
        values[0][4] = left[0][2] + right[0][1] + 1
        values[1][4] = left[1][2] + right[1][1]
        values[2][4] = left[2][2] + right[2][1] + 1

    max_contr = values[0].max()

    max_suppo = -1
    for i in range(5):
        if values[0][i] == max_contr and values[1][i] > max_suppo:
            max_suppo = values[1][i]

    max_oppos = -1
    for i in range(5):
        if values[0][i] == max_contr and values[2][i] > max_oppos:
            max_oppos = values[2][i]

    return max_contr, max_suppo, max_oppos


@njit('int64, int64[::3, ::5], int64[::3, ::5]',
      cache=True, nogil=True, boundscheck=False, parallel=False)
def calculate_max_given_condition(condition: int, left: np.array, right: np.array):  # parallel kills performance
    values = np.full(shape=(3, 9), fill_value=-1, dtype='int')

    if left[0][condition] > -1:
        # compare condition with all conditions
        for i in range(5):
            values[0][i] = left[0][condition] + right[0][i]
            values[1][i] = left[1][condition] + right[1][i]
            values[2][i] = left[2][condition] + right[2][i]

    if right[0][condition] > -1:
        col_id = 5
        # compare all conditions with condition
        for i in range(5):
            if i == condition:  # this comparison has already been made above
                continue

            values[0][col_id] = left[0][i] + right[0][condition]
            values[1][col_id] = left[1][i] + right[1][condition]
            values[2][col_id] = left[2][i] + right[2][condition]

            col_id += 1

    max_contr = values[0].max()

    max_suppo = -1
    for i in range(9):
        if values[0][i] == max_contr and values[1][i] > max_suppo:
            max_suppo = values[1][i]

    max_oppos = -1
    for i in range(9):
        if values[0][i] == max_contr and values[2][i] > max_oppos:
            max_oppos = values[2][i]

    return max_contr, max_suppo, max_oppos


@njit('int64[:, ::3, ::5], int64[:, ::3, ::5]',
      cache=True, nogil=True, boundscheck=False, parallel=False)
def combine_branches(left: np.array, right: np.array):
    assert left.shape == right.shape
    n_traits = left.shape[0]

    values = np.full(shape=left.shape, fill_value=-1, dtype='int')

    # selecting:values[<TRAITS>, <0: max; 1: supporting; 2: opposing>, <0: 11; 1: 10; 2: 01; 3: 00; 4: nf>]
    for trait_id in range(n_traits):  # prange kills performance
        for cond in range(4):  # prange kills performance
            # {"11": 0, "10": 1, "01": 2, "00": 3, "nf": 4}
            max_contr, max_suppo, max_oppos = calculate_max_given_condition(
                cond,
                left[trait_id, :, :],
                right[trait_id, :, :]
            )
            values[trait_id, 0, cond] = max_contr
            values[trait_id, 1, cond] = max_suppo
            values[trait_id, 2, cond] = max_oppos
        max_contr, max_suppo, max_oppos = calculate_max_nofree(
            left[trait_id, :, :],
            right[trait_id, :, :]
        )
        values[trait_id, 0, 4] = max_contr
        values[trait_id, 1, 4] = max_suppo
        values[trait_id, 2, 4] = max_oppos

    return values



import os
import logging
import pandas as pd
import numpy as np
import scipy.stats as ss

from .KeyValueStore import KeyValueStore
from .picking import pick
from .ScoaryTree import ScoaryTree

logger = logging.getLogger('scoary.permutations')


class ConfintStore(KeyValueStore):
    def create_db(self):
        self._create_db(
            columns={
                'tree': 'str',
                'n_pos_assoc': 'int',
                'n_permut': 'int',
                'confidence_interval': 'str'
            },
            pk_col='tree, n_pos_assoc, n_permut'
        )

    def get(self, tree: str, n_pos_assoc: int, n_permut: int):
        sql = f'SELECT confidence_interval FROM {self.table_name} WHERE tree = ? AND n_pos_assoc = ? AND n_permut = ?'
        res = self.cur.execute(
            sql,
            (tree, n_pos_assoc, n_permut,)
        ).fetchone()
        return np.frombuffer(res[0], dtype=float) if res is not None else None

    def set(self, tree: str, n_pos_assoc: int, n_permut: int, confidence_interval: [float]):
        confidence_interval = confidence_interval.tobytes()
        sql = f'INSERT OR IGNORE INTO {self.table_name} VALUES (?, ?, ?, ?)'
        self.cur.execute(
            sql, (tree, n_pos_assoc, n_permut, confidence_interval)
        )
        self.con.commit()


CONFINT_CACHE = ConfintStore(table_name='confint_cache', db_path=os.environ.get('CONFINT_DB', None))


def create_permuted_df(labels: [str], n_positive: int, n_permut: int, random_state: int = None):
    if random_state:
        np.random.seed(random_state)

    n_negative = len(labels) - n_positive
    arr = np.repeat(np.array([[1] * n_positive + [0] * n_negative]), n_permut, axis=0)

    # creates a copy -> slow
    arr = np.apply_along_axis(np.random.permutation, axis=1, arr=arr)

    return pd.DataFrame(arr, columns=labels)


def permute_picking(
        trait: str,
        tree: ScoaryTree,
        label_to_trait: pd.Series | dict,
        result_df: pd.DataFrame,
        genes_bool_df: pd.DataFrame,
        n_permut: int,
        random_state: int = None,
) -> np.array:
    if type(label_to_trait) is dict:
        label_to_trait = pd.Series(label_to_trait, dtype='boolean')
    n_tot = len(label_to_trait)
    n_pos = sum(label_to_trait)
    n_neg = n_tot - n_pos
    labels = label_to_trait.keys()

    n_reused = 0

    pvals = []
    for _, row in result_df.iterrows():
        label_to_gene = genes_bool_df.loc[row.Gene]
        unique_topology = tree.uniquify(label_to_gene)

        is_positively_correlated = row.supporting >= row.opposing
        estimator = (row.supporting if is_positively_correlated else row.opposing) / row.contrasting
        n_pos_assoc = n_pos if is_positively_correlated else n_neg

        permuted_estimators = CONFINT_CACHE.get(unique_topology, n_pos_assoc, n_permut)
        if permuted_estimators is None:
            permuted_df = create_permuted_df(
                labels=labels, n_positive=n_pos_assoc,
                n_permut=n_permut, random_state=random_state
            )
            max_contr, max_suppo, max_oppos = pick(
                tree=tree.to_list, label_to_trait_a=label_to_gene,
                trait_b_df=permuted_df, calc_pvals=False
            )

            permuted_estimators = max_suppo / max_contr
            CONFINT_CACHE.set(unique_topology, n_pos_assoc, n_permut, permuted_estimators)
        else:
            n_reused += 1

        pval = ((permuted_estimators >= estimator).sum() + 1) / (n_permut + 1)
        pvals.append(pval)

    logger.debug(f'{trait}: reused {n_reused} out of {len(result_df)}')

    return pvals



from .scoary import scoary
from .ScoaryTree import ScoaryTree
from .picking import pick, pick_single
from .permutations import permute_picking
from .utils import print_tree, get_version

__version__ = get_version()
__author__ = 'Thomas Roder'
__credits__ = ['Thomas Roder', 'Ola Brynildsrud']
__license__ = 'MIT'
__maintainer__ = 'Thomas Roder'
__email__ = 'roder.thomas@gmail.com'



import multiprocessing as mp

mp.set_start_method('spawn')


def init():
    mgr = mp.Manager()
    ns = mgr.Namespace()
    counter = mgr.Value('i', 0)
    lock = mgr.Lock()
    return mgr, ns, counter, lock



import os
import logging
import sqlite3


class KeyValueStore:
    table_name: str

    def __init__(self, table_name, db_path: str = None):
        self.table_name = table_name

        if db_path is None:
            if 'KEY_VALUE_STORE_DB' in os.environ:
                db_path = os.environ['KEY_VALUE_STORE_DB']
            else:
                db_path = os.path.expanduser('~/.cache/keyvaluestore.db')

        self._db_path = db_path
        self.con, self.cur = self.get_cur()
        self.create_db()

    def __str__(self):
        return f'KeyValueStore {self.table_name} ({self._db_path})'

    def get_cur(self):
        try:
            con = sqlite3.connect(self._db_path)
            cur = con.cursor()
        except Exception as e:
            logging.warning(f'Failed to connect to db: {self._db_path}')
            raise e
        return con, cur

    def __del__(self):
        try:
            self.cur.close()
            self.con.close()
        except Exception:
            pass

    def create_db(self):
        raise NotImplementedError(f'Users of the abstract class {self.__class__} must implement this function!')

    @staticmethod
    def list_to_string(l) -> str:
        return ', '.join(f"'{e}'" for e in l)

    @staticmethod
    def list_to_string_bracket(l):
        return ', '.join(f"('{e}')" for e in l)

    def _create_db(self, columns: {str: str}, pk_col: str):
        columns = ', '.join(f'{col_name} {col_type}' for col_name, col_type in columns.items())
        sql = f'''
            CREATE TABLE IF NOT EXISTS {self.table_name} (
                {columns},
                PRIMARY KEY ({pk_col})
            );
        '''
        try:
            self.cur.execute(sql)
        except sqlite3.OperationalError as e:
            logging.warning(f'Failed to run this SQL command on db {self._db_path}:\n{sql}')
            raise e

    def drop_db(self):
        self.cur.execute(f'''DROP TABLE {self.table_name}''')



import os
import logging
from collections import defaultdict
import json
import pandas as pd
from statsmodels.stats.multitest import multipletests
from fast_fisher.fast_fisher_numba import odds_ratio, test1t as fisher_exact_two_tailed
from queue import Empty

from .ScoaryTree import ScoaryTree
from .picking import pick
from .permutations import permute_picking
from .progressbar import print_progress
from .utils import setup_logging, AnalyzeTraitNamespace, fisher_id, grasp_namespace

logger = logging.getLogger('scoary.analyze_trait')


def worker(
        q,
        ns: AnalyzeTraitNamespace,
        result_container: {str: float | str | None},
        proc_id: int
):
    logger = setup_logging(
        logger=logging.getLogger('scoary'),
        path=f'{ns.outdir}/logs/scoary-2_proc{proc_id}.log',
        print_info=False,
        reset=True
    )
    logger.info(f'Setting up trait analysis worker {proc_id}')

    new_ns = grasp_namespace(AnalyzeTraitNamespace, ns)
    del ns

    local_result_container = {}

    while True:
        try:
            trait = q.get_nowait()
        except Empty:
            break  # completely done

        local_result_container[trait] = analyze_trait(trait, new_ns, proc_id)
        q.task_done()

    result_container.update(local_result_container)


def analyze_trait(trait: str, ns: AnalyzeTraitNamespace, proc_id: int = None) -> dict | str | None:
    logger.debug(f'Analyzing {trait=}')
    with ns.lock:
        ns.counter.value += 1
        message = trait if proc_id is None else f'P{proc_id} | {trait}'
        print_progress(
            ns.counter.value, len(ns.traits_df.columns),
            message=message, start_time=ns.start_time, message_width=25
        )

    if trait in ns.duplication_df:
        logger.debug(f'Duplicated trait: {trait} -> {ns.duplication_df[trait]}')
        save_duplicated_result(trait, ns)
        return ns.duplication_df[trait]

    trait_series = ns.traits_df[trait].dropna()
    labels = set(trait_series.index)

    if ns.all_labels == labels:
        pruned_tree = ns.tree
    else:
        pruned_tree = ns.tree.prune(labels=labels)

    result_df = init_result_df(ns.genes_bool_df, trait_series)
    test_df = create_test_df(result_df)
    test_df = add_odds_ratio(test_df)
    test_df = perform_multiple_testing_correction(
        test_df, col='fisher_p',
        method=ns.mt_f_method, cutoff=ns.mt_f_cutoff, is_sorted=True
    )

    if len(test_df) == 0:
        logger.info(f'Found 0 genes for {trait=} '
                    f'after {ns.mt_f_method}:{ns.mt_f_cutoff} filtration')
        return None

    result_df = pd.merge(
        test_df, result_df, how="left", on='__contingency_table__',
        copy=False  # for performance
    )

    result_df.attrs.update(test_df.attrs)

    if not ns.pairwise:
        result_df.attrs['best_fisher_p'] = result_df['fisher_p'].min()
        result_df.attrs['best_fisher_q'] = result_df['fisher_q'].min()
    else:
        result_df = pair_picking(
            result_df,
            significant_genes_df=ns.genes_bool_df.loc[result_df.Gene],
            tree=pruned_tree,
            label_to_trait=trait_series
        )

        if ns.worst_cutoff:
            if not (result_df['worst'] <= ns.worst_cutoff).any():
                logger.info(f'Found 0 genes for {trait=} '
                            f'after worst_cutoff={ns.worst_cutoff} filtration')
                return None

        if ns.max_genes:
            if len(result_df) > ns.max_genes:
                logger.info(f'Found too {len(result_df)} genes for {trait=} '
                            f'keeping only {ns.max_genes} with best Fisher\'s test.')
                result_df.attrs['max_genes'] = f'Trimmed {len(result_df)} genes to {ns.max_genes}.'
                result_df = result_df.iloc[:ns.max_genes]

        if ns.n_permut:
            result_df['empirical_p'] = permute_picking(
                trait=trait,
                result_df=result_df,
                tree=pruned_tree,
                label_to_trait=trait_series,
                n_permut=ns.n_permut,
                random_state=ns.random_state,
                genes_bool_df=ns.genes_bool_df
            )

            result_df['fq*ep'] = result_df['fisher_q'] * result_df['empirical_p']
            result_df.sort_values(by='fq*ep', inplace=True)

            result_df.attrs['best_fisher_p'] = result_df['fisher_p'][0]
            result_df.attrs['best_fisher_q'] = result_df['fisher_q'][0]
            result_df.attrs['best_empirical_p'] = result_df['empirical_p'][0]
            result_df.attrs['best_fq*ep'] = result_df['fq*ep'][0]

    save_result_df(trait, ns, result_df)

    # return minimal pvalues
    return result_df.attrs


def _save_trait(trait: str, ns: AnalyzeTraitNamespace):
    trait_df = pd.DataFrame(index=ns.traits_df.index)
    trait_df['class'] = ns.traits_df[trait]
    if ns.numeric_df is not None:
        trait_df['value'] = ns.numeric_df[trait]
    trait_df.index.name = 'isolate'
    trait_df.to_csv(f'{ns.outdir}/traits/{trait}/values.tsv', sep='\t')


def save_result_df(trait: str, ns: AnalyzeTraitNamespace, result_df: pd.DataFrame):
    os.makedirs(f'{ns.outdir}/traits/{trait}')

    # add annotations
    if ns.gene_info_df is None:
        additional_columns = []
    else:
        additional_columns = ns.gene_info_df.columns.to_list()
        result_df = result_df.merge(ns.gene_info_df, left_on='Gene', right_index=True, how='left', copy=False)

    # reorder columns
    col_order = ['Gene'] + \
                additional_columns + \
                ['g+t+', 'g+t-', 'g-t+', 'g-t-',
                 'sensitivity', 'specificity', 'odds_ratio',
                 'fisher_p', 'fisher_q', 'empirical_p', 'fq*ep',
                 'contrasting', 'supporting', 'opposing', 'best', 'worst']
    result_df = result_df[[col for col in col_order if col in result_df.columns]]

    result_df.to_csv(f'{ns.outdir}/traits/{trait}/result.tsv', sep='\t', index=False)

    binarization_info = ns.traits_df.attrs['binarization_info']
    if type(binarization_info) is str:
        binarization_info = defaultdict(lambda: 'none')

    with open(f'{ns.outdir}/traits/{trait}/meta.json', 'w') as f:
        meta_data = {
            'genes-content-type': ns.genes_orig_df.attrs['content_type'],
            'binarization-method': ns.traits_df.attrs['binarization_method'],
            'binarization-info': binarization_info[trait]
        }
        # add trait info
        if ns.trait_info_df is not None:
            try:
                info = ns.trait_info_df.loc[trait].to_dict()
                meta_data['info'] = {k: v for k, v in info.items() if not pd.isna(v)}
            except KeyError:
                pass

        json.dump(meta_data, f, indent=4, allow_nan=False)

    coverage_matrix = ns.genes_orig_df[ns.genes_orig_df.index.isin(result_df.Gene)].T
    coverage_matrix.index.name = 'Isolate'
    coverage_matrix.to_csv(f'{ns.outdir}/traits/{trait}/coverage-matrix.tsv', sep='\t')
    _save_trait(trait, ns)


def save_duplicated_result(trait: str, ns: AnalyzeTraitNamespace):
    # todo: might create symlinks traits that didn't make it through filtering!
    os.makedirs(f'{ns.outdir}/traits/{trait}')

    # use data from previous duplicate
    ref_trait = ns.duplication_df[trait]
    for f in ['result.tsv', 'meta.json', 'coverage-matrix.tsv']:
        os.symlink(src=f'../{ref_trait}/{f}', dst=f'{ns.outdir}/traits/{trait}/{f}')

    # create values.tsv only if numeric trait
    if ns.numeric_df is None:
        os.symlink(src=f'../{ref_trait}/values.tsv', dst=f'{ns.outdir}/traits/{trait}/values.tsv')
    else:
        _save_trait(trait, ns)


def init_result_df(genes_bool_df: pd.DataFrame, trait_series: pd.Series) -> pd.DataFrame:
    """
    Create result_df with index=strains and columns=[g+t+, g+t-, g-t+, g-t-, __contingency_table__]

    :param genes_bool_df: DataFrame (dtype: bool); columns: strains; rows: genes
    :param trait_series: Boolean Series that indicates which isolates have the trait
    :return: result_df (DataFrame); columns: ['g+t+', 'g+t-', 'g-t+', 'g-t-', '__contingency_table__]; index: strains
    """
    assert trait_series.dtype == 'boolean', f'trait_series must be boolean pandas.Series!'
    assert not trait_series.hasnans, f'trait_series may not contain NANs!'
    # Preparation
    trait_pos = trait_series.index[trait_series]
    trait_neg = trait_series.index[~trait_series]
    n_pos = len(trait_pos)
    n_neg = len(trait_neg)
    n_tot = n_pos + n_neg
    assert n_tot == len(trait_series)

    # create result_df
    result_df = pd.DataFrame(index=genes_bool_df.index)
    result_df['g+t+'] = genes_bool_df[trait_pos].sum(axis=1)  # trait positive gene positive
    result_df['g+t-'] = genes_bool_df[trait_neg].sum(axis=1)  # trait negative gene positive
    result_df['g-t+'] = n_pos - result_df['g+t+']  # trait positive gene negative
    result_df['g-t-'] = n_neg - result_df['g+t-']  # trait negative gene negative

    # remove genes that are shared by none or all
    gene_sum = result_df['g+t+'] + result_df['g+t-']
    result_df = result_df[(gene_sum != 0) & (gene_sum != n_tot)]

    # Add contingency table, sensitivity and specificity
    result_df['__contingency_table__'] = [tuple(x) for x in result_df[['g+t+', 'g+t-', 'g-t+', 'g-t-']].to_numpy()]
    result_df['sensitivity'] = (result_df['g+t+'] / n_pos * 100) if n_pos else 0
    result_df['specificity'] = (result_df['g-t-'] / n_neg * 100) if n_neg else 0

    # Reset index so that Gene is its own column
    result_df.reset_index(inplace=True)

    return result_df


def create_test_df(result_df: pd.DataFrame, sort=True) -> pd.DataFrame:
    """
    Create test_df with index=__contingency_id__ and columns=[fisher_p]

    Reduce to unique contingency tables
    Add column: fisher_p

    :param result_df: DataFrame with column '__contingency_table__'
    :param sort: whether to sort the DataFrame by pvalue
    :return: test_df (DataFrame)
    """

    test_df = pd.DataFrame(result_df.__contingency_table__.unique(), columns=['__contingency_table__'])

    # add __fisher_unique_table__
    test_df['__fisher_unique_table__'] = test_df.__contingency_table__.apply(lambda table: fisher_id(*table))

    # calculate Fisher's exact test
    table_to_pval = {table: fisher_exact_two_tailed(*table) for table in test_df.__fisher_unique_table__.unique()}

    # add Fisher's exact test
    test_df['fisher_p'] = test_df.__fisher_unique_table__.apply(lambda table: table_to_pval[table])

    # remove fisher_identifier
    test_df.drop('__fisher_unique_table__', axis=1, inplace=True)

    if sort:
        # sort test_df by pvalue
        test_df.sort_values(by='fisher_p', inplace=True)

    return test_df


def add_odds_ratio(test_df: pd.DataFrame) -> pd.DataFrame:
    # add odds_ratio
    test_df['odds_ratio'] = test_df.__contingency_table__.apply(lambda table: odds_ratio(*table))
    return test_df


def perform_multiple_testing_correction(
        test_df: pd.DataFrame,
        method: str,
        cutoff: float,
        col: str,
        is_sorted=True
) -> (float, pd.DataFrame):
    assert 'fisher_p' in col
    new_col = col.replace('fisher_p', 'fisher_q')

    if method == 'native':
        reject = test_df[col] <= cutoff
        _, qval, _, _ = multipletests(pvals=test_df[col], alpha=1, method='bonferroni', is_sorted=is_sorted)
    else:
        reject, qval, alphac_sidak, alphac_bonf = multipletests(
            pvals=test_df[col],
            alpha=cutoff,
            method=method,
            is_sorted=is_sorted,
        )

    test_df[new_col] = qval
    test_df = test_df[reject]
    return test_df


def pair_picking(result_df: pd.DataFrame, significant_genes_df: pd.DataFrame, tree: ScoaryTree,
                 label_to_trait: pd.Series | dict) -> pd.DataFrame:
    """
    Required rows:
    - Gene

    Add columns:
    - Max_Pairwise_comparisons
    - Max_supporting_pairs
    - Max_opposing_pairs
    - Best_pairwise_comp_p
    - Worst_pairwise_comp_p
    """
    assert result_df.Gene.to_list() == list(significant_genes_df.index)

    max_contr, max_suppo, max_oppos, best, worst = pick(
        tree=tree.to_list, label_to_trait_a=label_to_trait,
        trait_b_df=significant_genes_df, calc_pvals=True
    )

    result_df['contrasting'] = max_contr
    result_df['supporting'] = max_suppo
    result_df['opposing'] = max_oppos
    result_df['best'] = best
    result_df['worst'] = worst

    return result_df



from re import compile

BRANCH_LENTGHS = compile(r':[0-9]+(.[0-9]+)?')


class NewickParserException(Exception):
    pass


def parse_newick(newick_string: str) -> []:
    """
    A simple function to parse Newick strings to list tree format.

    Example:
        >>> parse_newick('(A,(B,C))D;')
        [['A', ['B', 'C']], 'D']

    Limitations:
        - Only binary trees are supported
        - Distances are ignored
        - All labels must be named
        - NHX format not supported

    :param newick_string: Phylogenetic tree in newick format
    :return: Phylogenetic tree in list format
    """

    # strip and remove branch lengths
    newick_string = newick_string.strip()
    newick_string = BRANCH_LENTGHS.sub(string=newick_string, repl='')

    # sanity check
    if not newick_string.endswith(';'):
        raise NewickParserException(f'Newick string does not end in semicolon! {newick_string=}')

    def find_corresponding_closing(string: str) -> int:
        n_opening = 0
        n_closing = 0
        for i, char in enumerate(string):
            if char == '(':
                n_opening += 1
                continue
            if char == ')':
                n_closing += 1
                if n_closing == n_opening:
                    return i

        raise NewickParserException(f'Could not find corresponding closing bracket in {string=}! {newick_string=}')

    def split_node(string: str) -> (str, str):
        if ',' in string:
            first_comma = string.index(',')
            if '(' in string:
                first_bracket = string.index('(')
                if first_bracket < first_comma:
                    return string[0:first_bracket], string[first_bracket:]

            return string[0:first_comma], string[first_comma + 1:]
        else:
            if '(' not in string:
                raise NewickParserException(f'Could not find separators "," or "(" in {string=}! {newick_string=}')
            first_bracket = string.index('(')
            return string[0:first_bracket], string[first_bracket:]

    def parse_leaf(string: str) -> str:
        string = string.strip('"\' ')
        if len(string) == 0:
            raise NewickParserException(f'Leaf with no label: {string=}! {newick_string=}')
        return string

    def parse_recursive(string: str) -> list | str:
        string = string.strip()

        # leaf
        if not ',' in string and not '(' in string and not ')' in string:
            return parse_leaf(string)

        # remove enclosing brackets
        if string.startswith('(') and string.endswith(')') \
                and find_corresponding_closing(string) == len(string) - 1:
            string = string[1:-1]

        # parse node
        if string.startswith('('):
            closing_idx = find_corresponding_closing(string)
            left = string[1:closing_idx]
            right = string[closing_idx + 1:].lstrip(',')
        else:
            left, right = split_node(string)

        left = parse_recursive(left)
        right = parse_recursive(right)

        return [left, right]

    return parse_recursive(newick_string[:-1])  # remove semicolon



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scoary']

package_data = \
{'': ['*'], 'scoary': ['templates/*']}

install_requires = \
['fast-fisher>=0.0.4,<0.0.5',
 'fire>=0.4.0,<0.5.0',
 'matplotlib>=3.5.2,<4.0.0',
 'numba>=0.55.1,<0.56.0',
 'pandas>=1.4.0,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'statsmodels>=0.13.2,<0.14.0']

entry_points = \
{'console_scripts': ['scoary2 = scoary.scoary:main']}

setup_kwargs = {
    'name': 'scoary-2',
    'version': '0.0.11',
    'description': 'Scoary2: Associate genes with traits!',
    'long_description': 'None',
    'author': 'MrTomRod',
    'author_email': 'roder.thomas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



class PhyloTree:
    """
    A class that represents a binary tree. Phylotrees can be nested.
    They can also contain tips at left, right or both nodes. Stores the
    max number of paths under each of the following 5 conditions:
    1. Free path to AB.
    2. Free path to Ab.
    3. Free path to aB.
    4. Free path to ab.
    5. No free path.
    """

    def __init__(
            self,
            leftnode,  # ['strainA', 'strainB'] or 'strainA'
            rightnode,  # ['strainA', 'strainB'] or 'strainA'
            GTC  # {'strain1': 'ab', 'strain2': 'Ab', 'strain3': 'ab'}
    ):
        """
        Constructs a phylotree and links it to its left and right nodes
        """
        # First check if left and right are tips (type = str). Then we
        # need to look up the gene-trait combination value at that node.
        # Also check if left and right nodes are PhyloTrees. If not,
        # recursively create PhyloTrees.

        if type(leftnode) is str:
            self.leftnode = Tip(GTC[leftnode], leftnode)
        elif isinstance(leftnode, PhyloTree):
            self.leftnode = leftnode
        else:
            self.leftnode = PhyloTree(leftnode=leftnode[0],
                                      rightnode=leftnode[1],
                                      GTC=GTC)

        if type(rightnode) is str:
            self.rightnode = Tip(GTC[rightnode], rightnode)
        elif isinstance(rightnode, PhyloTree):
            self.rightnode = rightnode
        else:
            self.rightnode = PhyloTree(leftnode=rightnode[0],
                                       rightnode=rightnode[1],
                                       GTC=GTC)

        # Initialize the max number of paths. Set to -1 meaning they
        # cannot be reached
        self.maxvalues = \
            {"AB": -1, "Ab": -1, "aB": -1, "ab": -1, "0": -1}
        self.max_propairs = \
            {"AB": -1, "Ab": -1, "aB": -1, "ab": -1, "0": -1}
        self.max_antipairs = \
            {"AB": -1, "Ab": -1, "aB": -1, "ab": -1, "0": -1}
        self.calculate_max()
        self.max_contrasting_pairs = max(self.maxvalues.values())
        self.max_contrasting_propairs = max(self.max_propairs.values())
        self.max_contrasting_antipairs = max(self.max_antipairs.values())

    def __str__(self):
        return f'({self.leftnode}, {self.rightnode})'

    def _to_newick(self):
        return f'({self.leftnode._to_newick()}, {self.rightnode._to_newick()})'

    def to_newick(self):
        return f'({self.leftnode._to_newick()}, {self.rightnode._to_newick()});'

    def to_label_to_gene(self):
        return self.leftnode.to_label_to_gene() | self.rightnode.to_label_to_gene()

    def to_label_to_trait(self):
        return self.leftnode.to_label_to_trait() | self.rightnode.to_label_to_trait()

    def calculate_max(self):
        """
        A method for calculating the max number of pairings under the 5
        conditions
        """

        for condition in ["AB", "Ab", "aB", "ab"]:
            pairings = self.calculate_max_condition(condition)  # {"Total": int, "Pro": int, "Anti": int}
            self.maxvalues[condition] = pairings["Total"]
            self.max_propairs[condition] = pairings["Pro"]
            self.max_antipairs[condition] = pairings["Anti"]
        pairings = self.calculate_max_nofree()
        self.maxvalues["0"] = pairings["Total"]
        self.max_propairs["0"] = pairings["Pro"]
        self.max_antipairs["0"] = pairings["Anti"]

    def calculate_max_condition(self, condition):
        """
        When passed for example 'AB', finds out the 9 distinct
        conditions and calculates the max. Here, the possibilities are
        illustrated when passed 'AB' (eg. a free path to AB exists):

        Left        Right
        AB          No free
        AB          Ab
        AB          aB
        AB          ab
        AB          AB
        No free     AB
        Ab          AB
        aB          AB
        ab          AB

        """
        Possible_conditions = {"AB", "Ab", "aB", "ab"}
        Possible_conditions.remove(condition)
        # Now we have a list of the elements that are NOT condition:
        other_conditions = list(Possible_conditions)
        max_pairs_1 = -1
        max_pairs_2 = -1
        max_pairs_3 = -1
        max_pairs_4 = -1
        max_pairs_5 = -1
        max_pairs_6 = -1
        max_pairs_7 = -1
        max_pairs_8 = -1
        max_pairs_9 = -1

        max_propairs_1 = -1
        max_propairs_2 = -1
        max_propairs_3 = -1
        max_propairs_4 = -1
        max_propairs_5 = -1
        max_propairs_6 = -1
        max_propairs_7 = -1
        max_propairs_8 = -1
        max_propairs_9 = -1

        max_antipairs_1 = -1
        max_antipairs_2 = -1
        max_antipairs_3 = -1
        max_antipairs_4 = -1
        max_antipairs_5 = -1
        max_antipairs_6 = -1
        max_antipairs_7 = -1
        max_antipairs_8 = -1
        max_antipairs_9 = -1

        l_maxvals = self.leftnode.maxvalues
        r_maxvals = self.rightnode.maxvalues

        l_maxpros = self.leftnode.max_propairs
        r_maxpros = self.rightnode.max_propairs

        l_maxantis = self.leftnode.max_antipairs
        r_maxantis = self.rightnode.max_antipairs

        # special case: comparison with nonfree
        if l_maxvals[condition] > -1 and r_maxvals["0"] > -1:
            max_pairs_1 = l_maxvals[condition] + r_maxvals["0"]
            max_propairs_1 = l_maxpros[condition] + r_maxpros["0"]
            max_antipairs_1 = l_maxantis[condition] + r_maxantis["0"]

        # comparison with all others
        if l_maxvals[condition] > -1 and r_maxvals[other_conditions[0]] > -1:
            max_pairs_2 = l_maxvals[condition] + r_maxvals[other_conditions[0]]
            max_propairs_2 = l_maxpros[condition] + r_maxpros[other_conditions[0]]
            max_antipairs_2 = l_maxantis[condition] + r_maxantis[other_conditions[0]]

        if l_maxvals[condition] > -1 and r_maxvals[other_conditions[1]] > -1:
            max_pairs_3 = l_maxvals[condition] + r_maxvals[other_conditions[1]]
            max_propairs_3 = l_maxpros[condition] + r_maxpros[other_conditions[1]]
            max_antipairs_3 = l_maxantis[condition] + r_maxantis[other_conditions[1]]

        if l_maxvals[condition] > -1 and r_maxvals[other_conditions[2]] > -1:
            max_pairs_4 = l_maxvals[condition] + r_maxvals[other_conditions[2]]
            max_propairs_4 = l_maxpros[condition] + r_maxpros[other_conditions[2]]
            max_antipairs_4 = l_maxantis[condition] + r_maxantis[other_conditions[2]]

        # special case: comparison with self
        if l_maxvals[condition] > -1 and r_maxvals[condition] > -1:
            max_pairs_5 = l_maxvals[condition] + r_maxvals[condition]
            max_propairs_5 = l_maxpros[condition] + r_maxpros[condition]
            max_antipairs_5 = l_maxantis[condition] + r_maxantis[condition]

        # comparison with all others
        if l_maxvals["0"] > -1 and r_maxvals[condition] > -1:
            max_pairs_6 = l_maxvals["0"] + r_maxvals[condition]
            max_propairs_6 = l_maxpros["0"] + r_maxpros[condition]
            max_antipairs_6 = l_maxantis["0"] + r_maxantis[condition]

        if l_maxvals[other_conditions[0]] > -1 and r_maxvals[condition] > -1:
            max_pairs_7 = l_maxvals[other_conditions[0]] + r_maxvals[condition]
            max_propairs_7 = l_maxpros[other_conditions[0]] + r_maxpros[condition]
            max_antipairs_7 = l_maxantis[other_conditions[0]] + r_maxantis[condition]

        if l_maxvals[other_conditions[1]] > -1 and r_maxvals[condition] > -1:
            max_pairs_8 = l_maxvals[other_conditions[1]] + r_maxvals[condition]
            max_propairs_8 = l_maxpros[other_conditions[1]] + r_maxpros[condition]
            max_antipairs_8 = l_maxantis[other_conditions[1]] + r_maxantis[condition]

        if l_maxvals[other_conditions[2]] > -1 and r_maxvals[condition] > -1:
            max_pairs_9 = l_maxvals[other_conditions[2]] + r_maxvals[condition]
            max_propairs_9 = l_maxpros[other_conditions[2]] + r_maxpros[condition]
            max_antipairs_9 = l_maxantis[other_conditions[2]] + r_maxantis[condition]

        max_pairs = max(max_pairs_1, max_pairs_2, max_pairs_3, max_pairs_4, max_pairs_5,
                        max_pairs_6, max_pairs_7, max_pairs_8, max_pairs_9)

        # Calculate maximum number of propairs, given a maxmimum number
        # of pairs
        max_propairs = -1
        if max_pairs == max_pairs_1:
            max_propairs = max(max_propairs, max_propairs_1)
        if max_pairs == max_pairs_2:
            max_propairs = max(max_propairs, max_propairs_2)
        if max_pairs == max_pairs_3:
            max_propairs = max(max_propairs, max_propairs_3)
        if max_pairs == max_pairs_4:
            max_propairs = max(max_propairs, max_propairs_4)
        if max_pairs == max_pairs_5:
            max_propairs = max(max_propairs, max_propairs_5)
        if max_pairs == max_pairs_6:
            max_propairs = max(max_propairs, max_propairs_6)
        if max_pairs == max_pairs_7:
            max_propairs = max(max_propairs, max_propairs_7)
        if max_pairs == max_pairs_8:
            max_propairs = max(max_propairs, max_propairs_8)
        if max_pairs == max_pairs_9:
            max_propairs = max(max_propairs, max_propairs_9)

        # Calculate maximum number of antipairs, given a maxmimum number
        # of pairs
        max_antipairs = -1
        if max_pairs == max_pairs_1:
            max_antipairs = max(max_antipairs, max_antipairs_1)
        if max_pairs == max_pairs_2:
            max_antipairs = max(max_antipairs, max_antipairs_2)
        if max_pairs == max_pairs_3:
            max_antipairs = max(max_antipairs, max_antipairs_3)
        if max_pairs == max_pairs_4:
            max_antipairs = max(max_antipairs, max_antipairs_4)
        if max_pairs == max_pairs_5:
            max_antipairs = max(max_antipairs, max_antipairs_5)
        if max_pairs == max_pairs_6:
            max_antipairs = max(max_antipairs, max_antipairs_6)
        if max_pairs == max_pairs_7:
            max_antipairs = max(max_antipairs, max_antipairs_7)
        if max_pairs == max_pairs_8:
            max_antipairs = max(max_antipairs, max_antipairs_8)
        if max_pairs == max_pairs_9:
            max_antipairs = max(max_antipairs, max_antipairs_9)

        return {"Total": max_pairs, "Pro": max_propairs, "Anti": max_antipairs}

    def calculate_max_nofree(self):
        """
        Under the condition of no free paths, only 5 distinct
        possibilities exits. (No free paths requires either that there
        are no free paths in the left or right nodes, or that a new pair
        is formed across the root from the left and right nodes)

        Left        Right       Result
        No free     No free
        AB          ab          +1 pair (pro)
        ab          AB          +1 pair (pro)
        Ab          aB          +1 pair (anti)
        aB          Ab          +1 pair (anti)
        """
        # No free pairs in either:
        max_pairs_nofree = -1
        max_pairs_1100 = -1
        max_pairs_0011 = -1
        max_pairs_1001 = -1
        max_pairs_0110 = -1

        max_propairs_nofree = -1
        max_propairs_1100 = -1
        max_propairs_0011 = -1
        max_propairs_1001 = -1
        max_propairs_0110 = -1

        max_antipairs_nofree = -1
        max_antipairs_1100 = -1
        max_antipairs_0011 = -1
        max_antipairs_1001 = -1
        max_antipairs_0110 = -1

        l_maxvals = self.leftnode.maxvalues
        r_maxvals = self.rightnode.maxvalues

        l_maxpros = self.leftnode.max_propairs
        r_maxpros = self.rightnode.max_propairs

        l_maxantis = self.leftnode.max_antipairs
        r_maxantis = self.rightnode.max_antipairs

        if l_maxvals["0"] > -1 and r_maxvals["0"] > -1:
            max_pairs_nofree = l_maxvals["0"] + r_maxvals["0"]
            max_propairs_nofree = l_maxpros["0"] + r_maxpros["0"]
            max_antipairs_nofree = l_maxantis["0"] + r_maxantis["0"]

        if l_maxvals["AB"] > -1 and r_maxvals["ab"] > -1:
            max_pairs_1100 = l_maxvals["AB"] + r_maxvals["ab"] + 1
            max_propairs_1100 = l_maxpros["AB"] + r_maxpros["ab"] + 1
            max_antipairs_1100 = l_maxantis["AB"] + r_maxantis["ab"]

        if l_maxvals["ab"] > -1 and r_maxvals["AB"] > -1:
            max_pairs_0011 = l_maxvals["ab"] + r_maxvals["AB"] + 1
            max_propairs_0011 = l_maxpros["ab"] + r_maxpros["AB"] + 1
            max_antipairs_0011 = l_maxantis["ab"] + r_maxantis["AB"]

        if l_maxvals["Ab"] > -1 and r_maxvals["aB"] > -1:
            max_pairs_1001 = l_maxvals["Ab"] + r_maxvals["aB"] + 1
            max_propairs_1001 = l_maxpros["Ab"] + r_maxpros["aB"]
            max_antipairs_1001 = l_maxantis["Ab"] + r_maxantis["aB"] + 1

        if l_maxvals["aB"] > -1 and r_maxvals["Ab"] > -1:
            max_pairs_0110 = l_maxvals["aB"] + r_maxvals["Ab"] + 1
            max_propairs_0110 = l_maxpros["aB"] + r_maxpros["Ab"]
            max_antipairs_0110 = l_maxantis["aB"] + r_maxantis["Ab"] + 1

        max_pairs = max(max_pairs_nofree, max_pairs_1100,
                        max_pairs_0011, max_pairs_1001, max_pairs_0110)

        # Calculate max number of propairs
        max_propairs = -1  # Max_propairs can never go below -1
        if max_pairs == max_pairs_nofree:
            max_propairs = max(max_propairs, max_propairs_nofree)
        if max_pairs == max_pairs_1100:
            max_propairs = max(max_propairs, max_propairs_1100)
        if max_pairs == max_pairs_0011:
            max_propairs = max(max_propairs, max_propairs_0011)
        if max_pairs == max_pairs_1001:
            max_propairs = max(max_propairs, max_propairs_1001)
        if max_pairs == max_pairs_0110:
            max_propairs = max(max_propairs, max_propairs_0110)

        # Calculate max number of antipairs
        max_antipairs = -1  # Max_antipairs can never go below -1
        if max_pairs == max_pairs_nofree:
            max_antipairs = max(max_antipairs, max_antipairs_nofree)
        if max_pairs == max_pairs_1100:
            max_antipairs = max(max_antipairs, max_antipairs_1100)
        if max_pairs == max_pairs_0011:
            max_antipairs = max(max_antipairs, max_antipairs_0011)
        if max_pairs == max_pairs_1001:
            max_antipairs = max(max_antipairs, max_antipairs_1001)
        if max_pairs == max_pairs_0110:
            max_antipairs = max(max_antipairs, max_antipairs_0110)

        return {"Total": max_pairs, "Pro": max_propairs, "Anti": max_antipairs}


class Tip:
    """
    A class that references a single tip, which can only be AB, Ab, aB,
    ab, A- or a-
    """

    def __init__(self, tipvalue, tipname: str):
        """
        Sets up the tip
        """
        self.tipvalue = tipvalue  # AB
        self.tipname = tipname  # strainA
        self.maxvalues = {}
        for condition in ["AB", "Ab", "aB", "ab", "0"]:
            if condition == tipvalue:
                self.maxvalues[condition] = 0
            else:
                self.maxvalues[condition] = -1
        self.max_propairs = {k: v for (k, v) in self.maxvalues.items()}
        self.max_antipairs = {k: v for (k, v) in self.maxvalues.items()}

    def __str__(self):
        tv1, tv2 = self.tipvalue
        assert tv1 in 'aA' and tv2 in 'bB'
        tv1 = '1' if tv1 == 'A' else '0'
        tv2 = '1' if tv2 == 'B' else '0'
        return f'{tv1}{tv2}_{self.tipname}'

    def _to_newick(self):
        return self.tipname

    def to_label_to_gene(self):
        tv1, tv2 = self.tipvalue
        assert tv1 in 'aA' and tv2 in 'bB'
        return {self.tipname: tv1 == 'A'}

    def to_label_to_trait(self):
        tv1, tv2 = self.tipvalue
        assert tv1 in 'aA' and tv2 in 'bB'
        return {self.tipname: tv2 == 'B'}


def convert_upgma_to_phylotree(tree, GTC):
    """
    A method that converts the upgma tree (in nested list form) to a
    PhyloTree. It also needs the status (AB, Ab, aB or ab) of all
    strains that are to be included in the analysis. As of 1.6.0 this
    function is also called from permutations
    """

    # TRAVERSING TREE: For each binary division - go to left until hit
    # tip. Then go back
    MyPhyloTree = PhyloTree(leftnode=tree[0],
                            rightnode=tree[1],
                            GTC=GTC)

    return MyPhyloTree, {"Total": MyPhyloTree.max_contrasting_pairs,
                         "Pro": MyPhyloTree.max_contrasting_propairs,
                         "Anti": MyPhyloTree.max_contrasting_antipairs}


import random


def permute_gtc(GTC):
    """
    Returns a permutation of the gene-trait combination dic where trait
    labels have been swapped

    ISOLATE KEEP GENES, GET PERMUTED TRAITS. TOTAL NUMBER OF TRAITS REMAINS
    """
    # Set of trait labels to distribute
    trait_labels = [s[-1] for s in GTC.values()]
    # Shuffle list of labels
    random.shuffle(trait_labels)
    for isolate in GTC:
        # Assign the latter character of GTC to isolates but keep the
        # gene status intact
        GTC[isolate] = str(GTC[isolate][0] + trait_labels.pop())
    return GTC
